#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
MobileNet model definition for ImageNet classification.

MobileNet V1 is a general architecture and can be used for multiple use cases.

This specific version includes parameter options to generate a mobilenet version
compatible for Akida with:
    - overall architecture compatible with Akida (conv stride 2 replaced with
     max pool),
    - options to quantize weights and activations,
    - different initialization options.
"""

import warnings

from keras import Model, regularizers
from keras.utils.data_utils import get_file
from keras.layers import (Input, Reshape, Dropout, Activation, Rescaling)

from cnn2snn import load_quantized_model, quantize

from ..layer_blocks import conv_block, separable_conv_block

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/mobilenet/'


def _obtain_input_shape(input_shape, default_size, min_size, include_top):
    """Internal utility to compute/validate a model's input shape.

    Args:
        input_shape: either None (will return the default model input shape),
            or a user-provided shape to be validated.
        default_size: default input width/height for the model.
        min_size: minimum input width/height accepted by the model.
        include_top: whether the model is expected to
            be linked to a classifier via a Flatten layer.

    Returns:
        tuple of integers: input shape (may include None entries).

    Raises:
        ValueError: in case of invalid argument values.
    """
    if input_shape and len(input_shape) == 3:
        if input_shape[-1] not in {1, 3}:
            warnings.warn('This model usually expects 1 or 3 input channels. '
                          'However, it was passed an input_shape with ' +
                          str(input_shape[-1]) + ' input channels.')
        default_shape = (default_size, default_size, input_shape[-1])
    else:
        default_shape = (default_size, default_size, 3)

    if input_shape:
        if input_shape is not None:
            if len(input_shape) != 3:
                raise ValueError(
                    '`input_shape` must be a tuple of three integers.')
            if ((input_shape[0] is not None and input_shape[0] < min_size) or
                (input_shape[1] is not None and input_shape[1] < min_size)):
                raise ValueError('Input size must be at least ' +
                                 str(min_size) + 'x' + str(min_size) +
                                 '; got `input_shape=' + str(input_shape) + '`')
    else:
        if include_top:
            input_shape = default_shape
        else:
            input_shape = (None, None, 3)
    if include_top:
        if None in input_shape:
            raise ValueError('If `include_top` is True, '
                             'you should specify a static `input_shape`. '
                             'Got `input_shape=' + str(input_shape) + '`')
    return input_shape


def mobilenet_imagenet(input_shape=None,
                       alpha=1.0,
                       dropout=1e-3,
                       include_top=True,
                       pooling=None,
                       classes=1000,
                       use_stride2=False,
                       weight_quantization=0,
                       activ_quantization=0,
                       input_weight_quantization=None,
                       input_scaling=(128, -1)):
    """Instantiates the MobileNet architecture.

    Args:
        input_shape (tuple): optional shape tuple.
        alpha (float): controls the width of the model.

            * If `alpha` < 1.0, proportionally decreases the number of filters
              in each layer.
            * If `alpha` > 1.0, proportionally increases the number of filters
              in each layer.
            * If `alpha` = 1, default number of filters from the paper are used
              at each layer.
        dropout (float): dropout rate
        include_top (bool): whether to include the fully-connected
            layer at the top of the model.
        pooling (str): Optional pooling mode for feature extraction
            when `include_top` is `False`.

            * `None` means that the output of the model will be the 4D tensor
              output of the last convolutional block.
            * `avg` means that global average pooling will be applied to the
              output of the last convolutional block, and thus the output of the
              model will be a 2D tensor.
        classes (int): optional number of classes to classify images
            into, only to be specified if `include_top` is True.
        use_stride2 (bool): optional, replace max pooling operations by stride 2
            convolutions in layers separable 2, 4, 6 and 12.
        weight_quantization (int): sets all weights in the model to have
            a particular quantization bitwidth except for the weights in the
            first layer.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization: sets all activations in the model to have a
            particular activation quantization bitwidth.

            * '0' implements floating point 32-bit activations.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_weight_quantization: sets weight quantization in the first layer.
            Defaults to weight_quantization value.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_scaling (tuple, optional): scale factor and offset to apply to
            inputs. Defaults to (128, -1). Note that following Akida convention,
            the scale factor is an integer used as a divider.

    Returns:
        keras.Model: a Keras model for MobileNet/ImageNet.

    Raises:
        ValueError: in case of invalid input shape.
    """
    # check if overrides have been provided and override
    if input_weight_quantization is None:
        input_weight_quantization = weight_quantization

    # Define weight regularization, will apply to the first convolutional layer
    # and to all pointwise weights of separable convolutional layers.
    weight_regularizer = regularizers.l2(4e-5)

    # Define stride 2 or max pooling
    if use_stride2:
        sep_conv_pooling = None
        strides = 2
    else:
        sep_conv_pooling = 'max'
        strides = 1

    # Determine proper input shape and default size.
    if input_shape is None:
        default_size = 224
    else:
        rows = input_shape[0]
        cols = input_shape[1]

        if rows == cols and rows in [128, 160, 192, 224]:
            default_size = rows
        else:
            default_size = 224

    input_shape = _obtain_input_shape(input_shape,
                                      default_size=default_size,
                                      min_size=32,
                                      include_top=include_top)

    rows = input_shape[0]
    cols = input_shape[1]

    img_input = Input(shape=input_shape)

    if input_scaling is None:
        x = img_input
    else:
        scale, offset = input_scaling
        x = Rescaling(1. / scale, offset)(img_input)

    x = conv_block(x,
                   filters=int(32 * alpha),
                   name='conv_0',
                   kernel_size=(3, 3),
                   padding='same',
                   use_bias=False,
                   strides=2,
                   add_batchnorm=True,
                   add_activation=True,
                   kernel_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(64 * alpha),
                             name='separable_1',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(128 * alpha),
                             name='separable_2',
                             kernel_size=(3, 3),
                             padding='same',
                             pooling=sep_conv_pooling,
                             strides=strides,
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(128 * alpha),
                             name='separable_3',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(256 * alpha),
                             name='separable_4',
                             kernel_size=(3, 3),
                             padding='same',
                             pooling=sep_conv_pooling,
                             strides=strides,
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(256 * alpha),
                             name='separable_5',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_6',
                             kernel_size=(3, 3),
                             padding='same',
                             pooling=sep_conv_pooling,
                             strides=strides,
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_7',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_8',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_9',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_10',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_11',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(1024 * alpha),
                             name='separable_12',
                             kernel_size=(3, 3),
                             padding='same',
                             pooling=sep_conv_pooling,
                             strides=strides,
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    # Last separable layer with global pooling
    layer_pooling = 'global_avg' if include_top or pooling == 'avg' else None
    x = separable_conv_block(x,
                             filters=int(1024 * alpha),
                             name='separable_13',
                             kernel_size=(3, 3),
                             padding='same',
                             pooling=layer_pooling,
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    if include_top:
        shape = (1, 1, int(1024 * alpha))

        x = Reshape(shape, name='reshape_1')(x)
        x = Dropout(dropout, name='dropout')(x)

        x = separable_conv_block(x,
                                 filters=classes,
                                 name='separable_14',
                                 kernel_size=(3, 3),
                                 padding='same',
                                 use_bias=False,
                                 add_batchnorm=False,
                                 add_activation=False,
                                 pointwise_regularizer=weight_regularizer)
        act_function = 'softmax' if classes > 1 else 'sigmoid'
        x = Activation(act_function, name=f'act_{act_function}')(x)
        x = Reshape((classes,), name='reshape_2')(x)

    # Create model.
    model = Model(img_input,
                  x,
                  name='mobilenet_%0.2f_%s_%s' % (alpha, rows, classes))

    if ((weight_quantization != 0) or (activ_quantization != 0) or
        (input_weight_quantization != 0)):
        return quantize(model, weight_quantization, activ_quantization,
                        input_weight_quantization)
    return model


def mobilenet_imagenet_pretrained(alpha=1.0):
    """
    Helper method to retrieve a `mobilenet_imagenet` model that was trained on
    ImageNet dataset.

    Args:
        alpha (float): width of the model.

    Returns:
        keras.Model: a Keras Model instance.

    """
    if alpha == 1.0:
        model_name = 'mobilenet_imagenet_224_iq8_wq4_aq4.h5'
        file_hash = '7d509ed9156af1b8739537d4b16b64fc6ad7320e94638ec8e554c483f6464aaf'
    elif alpha == 0.5:
        model_name = 'mobilenet_imagenet_224_alpha_50_iq8_wq4_aq4.h5'
        file_hash = 'daa6cb88c0422b8218c6aec97925a790aa08c3d52c14d70aec16ddd73a42fd2b'
    elif alpha == 0.25:
        model_name = 'mobilenet_imagenet_224_alpha_25_iq8_wq4_aq4.h5'
        file_hash = 'f90f6730ab02279883e0ae372b4d853b01fcf7fe354f5103685c74f119654492'
    else:
        raise ValueError(
            f"Requested model with alpha={alpha} is not available.")

    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
TSE/MLP model definition for CSE-CIC-IDS-2018 classification.
"""

import os

from keras import Model
from keras.utils.data_utils import get_file

from cnn2snn import load_quantized_model, quantize

from .preprocessing import load_labels_mapping, load_columns_type
from ..tabular_data.tabular_data import create_tse, create_mlp

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/tse_mlp/'


def tse_mlp_cse2018(numerical_columns=None,
                    categorical_columns=None,
                    weight_quantization=0,
                    activ_quantization=0,
                    input_weight_quantization=None):
    """Instantiates a model composed of a trainable spike encoder and a
    multilayer perceptron for the tabular data example on CSE-CIC-IDS-2018
    dataset.

    Args:
        numerical_columns (list): list of numerical column names. Will use
            preprocessed data when set to `None`.
        categorical_columns (dict): dictionary of categorical column names
            indexing the list of vocabulary for the column. Will use
            preprocessed data when set to `None`.
        weight_quantization (int): sets MLP weights in the model to have a
            particular quantization bitwidth

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization: sets all activations in the model to have a
            particular activation quantization bitwidth.

            * '0' implements floating point 32-bit activations.
            * '1' through '8' implements n-bit weights where n is from 1-8 bits.
        input_weight_quantization: sets weight quantization in the TSE last
            layer. Defaults to weight_quantization value.

            * 'None' implements the same bitwidth as the other weights.
            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.

    Returns:
        keras.Model: a Keras model for tabular data on CSE-CIC-IDS-2018
    """
    # Overrides input weight quantization if None
    if input_weight_quantization is None:
        input_weight_quantization = weight_quantization

    labels_mappings = load_labels_mapping()
    n_classes = len(labels_mappings)

    if numerical_columns is None or categorical_columns is None:
        preprocessed_numerical_columns, preprocessed_categorical_columns = load_columns_type(
        )

    if numerical_columns is None:
        numerical_columns = preprocessed_numerical_columns
    if categorical_columns is None:
        categorical_columns = preprocessed_categorical_columns

    mlp_units_list = [128, 128]

    inputs, tse_output = create_tse(numerical_columns,
                                    categorical_columns,
                                    units=256)
    mlp_output = create_mlp(tse_output, mlp_units_list, n_classes)

    model = Model(inputs=inputs, outputs=mlp_output)

    if ((weight_quantization != 0) or (activ_quantization != 0) or
        (input_weight_quantization != 0)):
        return quantize(model, weight_quantization, activ_quantization,
                        input_weight_quantization, False)
    return model


def tse_mlp_cse2018_pretrained():
    """
    Helper method to retrieve a `tse_mlp_cse2018` model that was trained on
    CSE-CIC-IDS-2018 dataset.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'tse_mlp_cse2018_iq0_wq2_aq4.tar.gz'
    file_hash = 'bdaa71424f18ed6368c259c6c7be05f5e455e9d97d834e55cc6da59b9cb8a503'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models',
                          extract=True)
    file_path = os.path.join(os.path.dirname(model_path),
                             "tse_mlp_cse2018_iq0_wq2_aq4")
    return load_quantized_model(file_path)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Layers blocks definitions.
"""

from keras.layers import (BatchNormalization, ReLU, Conv2D, SeparableConv2D,
                          Dense, MaxPool2D, AvgPool2D, GlobalAvgPool2D)


def _add_pooling_layer(x, pooling_type, pool_size, padding, layer_base_name):
    """Add a pooling layer in the graph.

    From an input tensor 'x', the function returns the output tensor after
    a pooling layer defined by 'pooling_type'.

    Args:
        x (tf.Tensor): the input tensor
        pooling_type (str): type of pooling among the following: 'max',
            'avg' or 'global_avg'.
        pool_size (int or tuple of 2 integers): factors by which to
            downscale (vertical, horizontal). (2, 2) will halve the input in
            both spatial dimension. If only one integer is specified, the same
            window length will be used for both dimensions.
        padding (str): one of "valid" or "same" (case-insensitive).
        layer_base_name (str): base name for the pooling layer.

    Returns:
        tf.Tensor: an output tensor after pooling
    """
    if pooling_type == 'max':
        return MaxPool2D(pool_size=pool_size,
                         padding=padding,
                         name=layer_base_name + '_maxpool')(x)
    if pooling_type == 'avg':
        return AvgPool2D(pool_size=pool_size,
                         padding=padding,
                         name=layer_base_name + '_avgpool')(x)
    if pooling_type == 'global_avg':
        return GlobalAvgPool2D(name=layer_base_name + '_global_avg')(x)
    raise ValueError("'pooling_type' argument must be 'max', "
                     "'avg' or 'global_avg'.")


def conv_block(inputs,
               filters,
               kernel_size,
               pooling=None,
               pool_size=(2, 2),
               add_batchnorm=False,
               add_activation=True,
               **kwargs):
    """Adds a convolutional layer with optional layers in the following order:
    max pooling, batch normalization, activation.

    Args:
        inputs (tf.Tensor): input tensor of shape `(rows, cols, channels)`
        filters (int): the dimensionality of the output space
            (i.e. the number of output filters in the convolution).
        kernel_size (int or tuple of 2 integers): specifying the
            height and width of the 2D convolution kernel.
            Can be a single integer to specify the same value for
            all spatial dimensions.
        pooling (str): add a pooling layer of type 'pooling' among the
            values 'max', 'avg', 'global_max' or 'global_avg', with pooling
            size set to pool_size. If 'None', no pooling will be added.
        pool_size (int or tuple of 2 integers): factors by which to
            downscale (vertical, horizontal). (2, 2) will halve the input in
            both spatial dimension. If only one integer is specified, the same
            window length will be used for both dimensions.
        add_batchnorm (bool): add a BatchNormalization layer
        add_activation (bool): add a ReLU layer
        **kwargs: arguments passed to the keras.Conv2D layer, such as
            strides, padding, use_bias, weight_regularizer, etc.

    Returns:
        tf.Tensor: output tensor of conv2D block.
    """
    if 'activation' in kwargs and kwargs['activation']:
        raise ValueError(
            "Keyword argument 'activation' in conv_block must be None.")
    if 'dilation_rate' in kwargs:
        if kwargs['dilation_rate'] != 1 and kwargs['dilation_rate'] != (1, 1):
            raise ValueError("Keyword argument 'dilation_rate' is not "
                             "supported in conv_block.")

    conv_layer = Conv2D(filters, kernel_size, **kwargs)
    x = conv_layer(inputs)

    if pooling:
        x = _add_pooling_layer(x, pooling, pool_size, conv_layer.padding,
                               conv_layer.name)

    if add_batchnorm:
        x = BatchNormalization(name=conv_layer.name + '_BN')(x)

    if add_activation:
        x = ReLU(6.0, name=conv_layer.name + '_relu')(x)

    return x


def separable_conv_block(inputs,
                         filters,
                         kernel_size,
                         pooling=None,
                         pool_size=(2, 2),
                         add_batchnorm=False,
                         add_activation=True,
                         **kwargs):
    """Adds a separable convolutional layer with optional layers in the
    following order: global average pooling, max pooling, batch normalization,
    activation.

    Args:
        inputs (tf.Tensor): input tensor of shape `(height, width, channels)`
        filters (int): the dimensionality of the output space
            (i.e. the number of output filters in the pointwise convolution).
        kernel_size (int or tuple of 2 integers): specifying the
            height and width of the 2D convolution window. Can be a single
            integer to specify the same value for all spatial dimensions.
        pooling (str): add a pooling layer of type 'pooling' among the
            values 'max', 'avg', 'global_max' or 'global_avg', with pooling
            size set to pool_size. If 'None', no pooling will be added.
        pool_size (int or tuple of 2 integers): factors by which to
            downscale (vertical, horizontal). (2, 2) will halve the input in
            both spatial dimension. If only one integer is specified, the same
            window length will be used for both dimensions.
        add_batchnorm (bool): add a BatchNormalization layer
        add_activation (bool): add a ReLU layer
        **kwargs: arguments passed to the keras.SeparableConv2D layer,
            such as strides, padding, use_bias, etc.

    Returns:
        tf.Tensor: output tensor of separable conv block.
    """
    if 'activation' in kwargs and kwargs['activation']:
        raise ValueError("Keyword argument 'activation' in separable_conv_block"
                         " must be None.")
    if 'dilation_rate' in kwargs:
        if kwargs['dilation_rate'] != 1 and kwargs['dilation_rate'] != (1, 1):
            raise ValueError("Keyword argument 'dilation_rate' is not "
                             "supported in separable_conv_block.")
    if 'depth_multiplier' in kwargs:
        if kwargs['depth_multiplier'] != 1:
            raise ValueError("Keyword argument 'depth_multiplier' is not "
                             "supported in separable_conv_block.")

    sep_conv_layer = SeparableConv2D(filters, kernel_size, **kwargs)
    x = sep_conv_layer(inputs)

    if pooling:
        x = _add_pooling_layer(x, pooling, pool_size, sep_conv_layer.padding,
                               sep_conv_layer.name)

    if add_batchnorm:
        x = BatchNormalization(name=sep_conv_layer.name + '_BN')(x)

    if add_activation:
        x = ReLU(6.0, name=sep_conv_layer.name + '_relu')(x)

    return x


def dense_block(inputs,
                units,
                add_batchnorm=False,
                add_activation=True,
                **kwargs):
    """Adds a dense layer with optional layers in the following order:
    batch normalization, activation.

    Args:
        inputs (tf.Tensor): Input tensor of shape `(rows, cols, channels)`
        units (int): dimensionality of the output space
        add_batchnorm (bool): add a BatchNormalization layer
        add_activation (bool): add a ReLU layer
        **kwargs: arguments passed to the Dense layer, such as
            use_bias, kernel_initializer, weight_regularizer, etc.

    Returns:
        tf.Tensor: output tensor of the dense block.
    """
    if 'activation' in kwargs and kwargs['activation']:
        raise ValueError("Keyword argument 'activation' in dense_block"
                         " must be None.")

    dense_layer = Dense(units, **kwargs)
    x = dense_layer(inputs)

    if add_batchnorm:
        x = BatchNormalization(name=dense_layer.name + '_BN')(x)

    if add_activation:
        x = ReLU(6.0, name=dense_layer.name + '_relu')(x)

    return x



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Common utility methods used by the training scripts.
"""

import argparse
import numpy as np

from keras.preprocessing.image import DirectoryIterator
from keras.optimizer_v2.adam import Adam
from keras.layers import Activation


def compile_model(model,
                  learning_rate=1e-3,
                  loss='categorical_crossentropy',
                  metrics=None):
    """ Compiles the model using Adam optimizer.

    Args:
        model (keras.Model): the model to compile
        learning_rate (float, optional): the learning rate. Defaults to 1e-3.
        loss (str or function, optional): the loss function. Defaults to
            'categorical_crossentropy'.
        metrics (list, optional): list of metrics to be evaluated during
            training and testing. Defaults to None.
    """
    if metrics is None:
        metrics = ['accuracy']

    model.compile(optimizer=Adam(learning_rate=learning_rate),
                  loss=loss,
                  metrics=metrics)


def freeze_model_before(model, freeze_before):
    """ Freezes the model before the given layer name.

    Args:
        model (keras.Model): the model to freeze
        freeze_before (str): name of the layer from which the model will not be
            frozen

    Raises:
        ValueError: if the provided layer name was not found in the model
    """
    if freeze_before is not None:
        trainable = False
        for layer in model.layers:
            if layer.name == freeze_before:
                trainable = True
            layer.trainable = trainable
        if not trainable:
            raise ValueError(f"No such layer {freeze_before} in model.")


def evaluate_model(model,
                   x,
                   y=None,
                   batch_size=None,
                   steps=None,
                   print_history=False):
    """ Evaluates model performances.

    Args:
        model (keras.Model): the model to evaluate
        x (tf.Dataset, np.array or generator): evaluation input data
        y (tf.Dataset, np.array or generator, optional): evaluation target
            data. Defaults to None.
        batch_size (int, optional): the batch size. Defaults to None.
        steps (int, optional): total number of steps before declaring the
            evaluation round finished. Defaults to None.
        print_history (bool, optional): either to print all history or just
            accuracy. Defaults to False.
    """
    history = model.evaluate(x=x,
                             y=y,
                             batch_size=batch_size,
                             steps=steps,
                             verbose=0)

    if print_history:
        print('Validation history: ', history)
    else:
        print('Validation accuracy: ', history[1])


def evaluate_akida_model(model, x, activation):
    """ Evaluates Akida model and return predictions and labels to compute
    accuracy.

    Args:
        model (akida.Model): the model to evaluate
        x (tf.Dataset, np.array or generator): evaluation input data
        activation (str): activation function to apply to potentials

    Returns:
        np.array, np.array: predictions and labels
    """

    # Initialize to None to allow different shapes depending on the caller
    labels = None
    pots = None

    # Datasets created as ImageDataGenerator will be looped over indefinitely,
    # they thus need a special treatment
    steps = None
    if isinstance(x, DirectoryIterator):
        steps = np.math.ceil(x.samples / x.batch_size)

    for batch, label_batch in x:
        if not isinstance(batch, np.ndarray):
            batch = batch.numpy()

        pots_batch = model.evaluate(batch.astype('uint8'))

        if labels is None:
            labels = label_batch
            pots = pots_batch.squeeze()
        else:
            labels = np.concatenate((labels, label_batch))
            pots = np.concatenate((pots, pots_batch.squeeze()))

        # End the for loop if the number of batches has reached the number of
        # steps
        if steps is not None and x.total_batches_seen == steps:
            break

    return Activation(activation)(pots), labels


def get_training_parser(batch_size,
                        tune=False,
                        freeze_before=False,
                        global_batch_size=True,
                        global_parser=None):
    """ Instantiates a base arguments parser for training scripts.

    The returned parser comes with train, tune and eval actions as subparsers
    (with default parameters) for which one can add arguments if required.

    Args:
        batch_size (int): batch size default value
        tune (bool, optional): either to add a tune action or not. Default to
            False.
        freeze_before (bool, optional): either to add a "--freeze_before"
            parameter in train and tune actions or not. Default to False.
        global_batch_size (bool, optional): either to add batch size as a
            global parser parameter or for train/tune actions only. Default to
            True.
        global_parser (argparse.ArgumentParser, optional): global parser with
            custom parameters. Defaults to None.

    Returns:
        argparse.ArgumentParser, argparse.ArgumentParser,
        argparse.ArgumentParser, argparse.ArgumentParser,
        argparse._SubParsersAction: main parser and train, tune and eval actions
            subparsers and the subparser object.
    """

    # Define a strictly positive int type for parameters
    def positive_int(value):
        ivalue = int(value)
        if ivalue <= 0:
            raise argparse.ArgumentTypeError(
                "%s is an invalid value, must be >0." % value)
        return ivalue

    parser = argparse.ArgumentParser()

    # Base arguments
    if global_parser is None:
        global_parser = argparse.ArgumentParser(add_help=False)

    global_parser.add_argument("-m",
                               "--model",
                               type=str,
                               required=True,
                               help="Model to load")
    if global_batch_size:
        global_parser.add_argument("-b",
                                   "--batch_size",
                                   type=positive_int,
                                   default=batch_size,
                                   help="The batch size")

    # Create an action subparser
    subparsers = parser.add_subparsers(help="Desired action to perform",
                                       dest="action")

    # Create parent subparser for arguments shared between train and tune
    # actions
    parent_parser = argparse.ArgumentParser(add_help=False,
                                            parents=[global_parser])
    parent_parser.add_argument("-s",
                               "--savemodel",
                               type=str,
                               default=None,
                               help="Save model with the specified name")
    parent_parser.add_argument("-e",
                               "--epochs",
                               type=positive_int,
                               required=True,
                               help="The number of training epochs")
    if not global_batch_size:
        parent_parser.add_argument("-b",
                                   "--batch_size",
                                   type=positive_int,
                                   default=batch_size,
                                   help="The batch size")
    if freeze_before:
        parent_parser.add_argument(
            "-fb",
            "--freeze_before",
            type=str,
            default=None,
            help="Freeze the layers of the model before this \
                                layer")

    # Subparsers based on parent
    train_parser = subparsers.add_parser("train",
                                         parents=[parent_parser],
                                         help="Train a Keras model")
    tune_parser = None
    if tune:
        tune_parser = subparsers.add_parser("tune",
                                            parents=[parent_parser],
                                            help="Tune a Keras model)")

    # Other subparsers
    eval_parser = subparsers.add_parser("eval",
                                        help="Evaluate a model",
                                        parents=[global_parser])

    eval_parser.add_argument("-ak",
                             "--akida",
                             action='store_true',
                             help="Converts to an Akida model and evaluate")

    return parser, train_parser, tune_parser, eval_parser, subparsers



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
akida_models main command-line interface

This package entry-point allows akida models to be instantiated and saved at a
specified location.

"""

import argparse

from .cifar10.model_ds_cnn import ds_cnn_cifar10
from .cifar10.model_vgg import vgg_cifar10
from .utk_face.model_vgg import vgg_utk_face
from .kws.model_ds_cnn import ds_cnn_kws
from .modelnet40.model_pointnet_plus import pointnet_plus_modelnet40
from .imagenet.model_mobilenet import mobilenet_imagenet
from .imagenet.model_vgg import vgg_imagenet
from .imagenet.model_akidanet import akidanet_imagenet
from .imagenet.model_akidanet_edge import akidanet_edge_imagenet
from .imagenet.model_mobilenet_edge import mobilenet_edge_imagenet
from .detection.model_yolo import yolo_base
from .dvs.model_convtiny_handy import convtiny_dvs_handy
from .dvs.model_convtiny_gesture import convtiny_dvs_gesture
from .cwru.model_convtiny import convtiny_cwru
from .mnist.model_gxnor import gxnor_mnist


def main():
    """ CLI entry point.

    Contains an argument parser with specific arguments depending on the model
    to be created. Complete arguments lists available using the -h or --help
    argument.

    """
    parser = argparse.ArgumentParser()
    sp = parser.add_subparsers(dest="action")
    c_parser = sp.add_parser("create", help="Create an akida Keras model")
    c_parser.add_argument("-s",
                          "--save_model",
                          type=str,
                          default=None,
                          help="The path/name to use to save the model")
    csp = c_parser.add_subparsers(dest="model",
                                  help="The type of model to be instantiated")
    csp.add_parser(
        "ds_cnn_cifar10",
        help="A Depthwise Separable CIFAR10 model inspired by MobileNet")
    csp.add_parser("vgg_cifar10", help="A VGG-like CIFAR10 model")
    csp.add_parser("vgg_utk_face", help="A VGG-like UTKFace model")
    csp.add_parser("convtiny_dvs_handy", help="A Convtiny DVS handy model")
    dvsh_parser = csp.add_parser("convtiny_dvs_handy",
                                 help="A Convtiny DVS model for akida")
    dvsh_parser.add_argument("-c",
                             "--classes",
                             type=int,
                             default=9,
                             help="The number of classes")
    csp.add_parser("convtiny_dvs_gesture", help="A Convtiny DVS gesture model")
    dvsg_parser = csp.add_parser("convtiny_dvs_gesture",
                                 help="A Convtiny DVS model for akida")
    dvsg_parser.add_argument("-c",
                             "--classes",
                             type=int,
                             default=9,
                             help="The number of classes")
    csp.add_parser(
        "ds_cnn_kws",
         help="A Depthwise Separable MobileNet-like model for the Keyword" \
              " Spotting example")
    mb_parser = csp.add_parser("mobilenet_imagenet",
                               help="A MobileNet V1 model for akida")
    image_sizes = [32, 64, 96, 128, 160, 192, 224]
    mb_parser.add_argument("-i",
                           "--image_size",
                           type=int,
                           default=224,
                           choices=image_sizes,
                           help="The square input image size")
    mb_parser.add_argument("-a",
                           "--alpha",
                           type=float,
                           default=1.0,
                           help="The width of the model")
    mb_parser.add_argument("-c",
                           "--classes",
                           type=int,
                           default=1000,
                           help="The number of classes")
    vgg_parser = csp.add_parser("vgg_imagenet", help="A VGG model for akida")
    vgg_parser.add_argument("-c",
                            "--classes",
                            type=int,
                            default=1000,
                            help="The number of classes")
    mbe_parser = csp.add_parser("mobilenet_imagenet_edge",
                                help="A MobileNet V1 model modified for akida \
                                edge learning")
    mbe_parser.add_argument("-bm",
                            "--base_model",
                            type=str,
                            required=True,
                            help="The base MobileNet model to use for edge \
                            adaptation.")
    mbe_parser.add_argument("-c",
                            "--classes",
                            type=int,
                            default=1000,
                            help="The number of edge learning classes")
    an_parser = csp.add_parser("akidanet_imagenet", help="An AkidaNet model")
    an_parser.add_argument("-i",
                           "--image_size",
                           type=int,
                           default=224,
                           choices=image_sizes,
                           help="The square input image size")
    an_parser.add_argument("-a",
                           "--alpha",
                           type=float,
                           default=1.0,
                           help="The width of the model")
    an_parser.add_argument("-c",
                           "--classes",
                           type=int,
                           default=1000,
                           help="The number of classes")
    ane_parser = csp.add_parser("akidanet_edge_imagenet",
                                help="An AkidaNet model modified for akida \
                                edge learning")
    ane_parser.add_argument("-bm",
                            "--base_model",
                            type=str,
                            required=True,
                            help="The base AkidaNet model to use for edge \
                            adaptation.")
    ane_parser.add_argument("-c",
                            "--classes",
                            type=int,
                            default=1000,
                            help="The number of edge learning classes")
    pnet_parser = csp.add_parser("pointnet_plus_modelnet40",
                                 help="A PointNet++ model for Akida")
    pnet_parser.add_argument("-c",
                             "--classes",
                             type=int,
                             default=40,
                             help="The number of classes")
    pnet_parser.add_argument("-alpha",
                             "--alpha",
                             type=float,
                             default=1.0,
                             help="Network filters multiplier (typically 0.5 , \
                             [1.0]")
    yl_parser = csp.add_parser("yolo_base", help="A YOLOv2 model for detection")
    yl_parser.add_argument("-c",
                           "--classes",
                           type=int,
                           default=1,
                           help="The number of classes")
    yl_parser.add_argument("-na",
                           "--number_anchors",
                           type=int,
                           default=5,
                           help="The number of anchor boxes")
    yl_parser.add_argument("-a",
                           "--alpha",
                           type=float,
                           default=0.5,
                           help="The width of the model")
    yl_parser.add_argument("-bw",
                           "--base_weights",
                           type=str,
                           default=None,
                           help="The base MobileNet weights to use for \
                            transfer learning.")
    csp.add_parser("convtiny_cwru", help="A Convtiny CWRU model for akida")
    csp.add_parser("gxnor_mnist", help="A GXNOR MNIST model for akida")
    args = parser.parse_args()
    if args.action == "create":
        # Instantiate the wished model
        if args.model == "ds_cnn_cifar10":
            model = ds_cnn_cifar10()
        elif args.model == "vgg_cifar10":
            model = vgg_cifar10()
        elif args.model == "vgg_utk_face":
            model = vgg_utk_face()
        elif args.model == "convtiny_dvs_handy":
            model = convtiny_dvs_handy(classes=args.classes)
        elif args.model == "convtiny_dvs_gesture":
            model = convtiny_dvs_gesture(classes=args.classes)
        elif args.model == "ds_cnn_kws":
            model = ds_cnn_kws()
        elif args.model == "pointnet_plus_modelnet40":
            model = pointnet_plus_modelnet40(alpha=args.alpha,
                                             classes=args.classes)
        elif args.model == "mobilenet_imagenet":
            input_shape = (args.image_size, args.image_size, 3)
            model = mobilenet_imagenet(input_shape,
                                       alpha=args.alpha,
                                       classes=args.classes)
        elif args.model == "vgg_imagenet":
            model = vgg_imagenet(classes=args.classes)
        elif args.model == "mobilenet_imagenet_edge":
            model = mobilenet_edge_imagenet(base_model=args.base_model,
                                            classes=args.classes)
        elif args.model == "akidanet_imagenet":
            input_shape = (args.image_size, args.image_size, 3)
            model = akidanet_imagenet(input_shape,
                                      alpha=args.alpha,
                                      classes=args.classes)
        elif args.model == "akidanet_edge_imagenet":
            model = akidanet_edge_imagenet(base_model=args.base_model,
                                           classes=args.classes)
        elif args.model == "yolo_base":
            model = yolo_base(classes=args.classes,
                              nb_box=args.number_anchors,
                              alpha=args.alpha)
            if args.base_weights is not None:
                model.load_weights(args.base_weights, by_name=True)
        elif args.model == "convtiny_cwru":
            model = convtiny_cwru()
        elif args.model == "gxnor_mnist":
            model = gxnor_mnist()
        # No need for default behaviour as the command-line parser only accepts
        # valid model types
        model_path = args.save_model
        if model_path is None:
            model_path = \
                f"{model.name}.h5" if args.model in ("mobilenet_imagenet", "mobilenet_imagenet_edge") else f"{args.model}.h5"

        # If needed, add the extension
        if not model_path.endswith(".h5"):
            model_path = f"{model_path}.h5"
        model.save(model_path, include_optimizer=False)
        print(f"Model saved as {model_path}")



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
GXNOR model definition for MNIST classification.
"""

from keras import Model
from keras.layers import Input, Flatten, Rescaling
from keras.utils.data_utils import get_file

from cnn2snn import quantize, load_quantized_model

from ..layer_blocks import conv_block, dense_block

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/gxnor/'


def gxnor_mnist(weight_quantization=0,
                activ_quantization=0,
                input_weight_quantization=None):
    """ Instantiates a Keras GXNOR model with an additional dense layer to make
    better classification.

    The paper describing the original model can be found `here
    <https://www.sciencedirect.com/science/article/pii/S0893608018300108>`_.

    Args:
        weight_quantization (int, optional): sets all weights in the model to
            have a particular quantization bitwidth except for the weights in
            the first layer. Defaults to 0.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization (int, optional): sets all activations in the model to
            have a particular activation quantization bitwidth. Defaults to 0.

            * '0' implements floating point 32-bit activations.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_weight_quantization(int, optional): sets weight quantization in
            the first layer. Defaults to weight_quantization value.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.

    Returns:
        keras.Model: a Keras model for GXNOR/MNIST
    """

    # check if overrides have been provided and override
    if input_weight_quantization is None:
        input_weight_quantization = weight_quantization

    img_input = Input(shape=(28, 28, 1))
    x = Rescaling(1. / 255)(img_input)

    # Block 1
    x = conv_block(x,
                   filters=32,
                   name='block1_conv1',
                   kernel_size=(5, 5),
                   padding='same',
                   add_batchnorm=True,
                   add_activation=True,
                   pooling='max',
                   pool_size=(2, 2))

    # Block 2
    x = conv_block(x,
                   filters=64,
                   name='block2_conv1',
                   kernel_size=(5, 5),
                   padding='same',
                   add_batchnorm=True,
                   add_activation=True,
                   pooling='max',
                   pool_size=(2, 2))

    # Classification block
    x = Flatten(name='flatten')(x)
    x = dense_block(x,
                    units=512,
                    name='fc1',
                    add_batchnorm=True,
                    add_activation=True)
    x = dense_block(x,
                    units=10,
                    name='predictions',
                    add_batchnorm=True,
                    add_activation=False)

    # Create model
    model = Model(img_input, x, name='gxnor_mnist')

    if ((weight_quantization != 0) or (activ_quantization != 0) or
        (input_weight_quantization != 0)):
        return quantize(model, weight_quantization, activ_quantization,
                        input_weight_quantization)
    return model


def gxnor_mnist_pretrained():
    """ Helper method to retrieve a `gxnor_mnist` model that was trained on
    MNIST dataset.

    This model was trained with the distillation knowledge method, using
    the EfficientNet model from `this repository
    <https://github.com/EscVM/Efficient-CapsNet>`_ and the `Distiller` class from
    the knowledge distillation toolkit (`akida_models.distiller`).

    The float training was done for 30 epochs with a learning rate of 1e-4
    After that we gradually quantize the model from:
    8-4-4 --> 4-4-4 --> 4-4-2 --> 2-2-2 --> 2-2-1 tuning the model at each step
    with the same distillation training method for 5 epochs and a learning rate
    of 5e-5.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'gxnor_mnist_iq2_wq2_aq1.h5'
    file_hash = '83405feaab26ed62fcf20f50eb10b93ad696f15ed6a37b54cc97ea6ed996ef36'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
 DVS DataGenerator class.

 This script is used in dvs_train.py to generate images from DVS data from NPY
    files. It's mandatory because Tensorflow doesn't handle events data.
"""

import random
import numpy as np


class DVS_DataGenerator:
    """ Data generator for handling event-based camera data.

    Args:
        dataset (array): array of data stored in a dict.
        inversion (bool, optional): flag to flip x/y coords.
        packets_fixed_by (str, optional): value used to fix packet length, one
            of 'duration' or (number of) 'events'. Defaults to 'duration'.
        subpacket_length (int, optional): subpacket length (milliseconds if
            packets_fixed_by='duration', event count if 'events'). Defaults to
            100.
        subpackets_per_packet (int, optional): number of sub-packets (~ time
            bins) to concatenate to generate packets. Defaults to 1.
        camera_dims (tuple of ints, optional): dimensions of the raw data from
            the camera. Defaults to (128, 128, 2).
        include_polarity (bool, optional): keep event polarity. If False, On and
            Off events will be treated as if identical. Defaults to True.
        allow_duplicate_events (bool, optional): allow events to be
            superimposed. If False, only a single event will be allowed per
            pixel for each timebin (subpacket) and polarity. Default to False.
        downscale (int, optional): downscaling of inputs, implemented as integer
            division of incoming coordinate values. Defaults to 1.

    Returns:
        tuple: two numpy arrays with events and labels.
    """

    def __init__(self,
                 dataset,
                 inversion=False,
                 packets_fixed_by='duration',
                 subpacket_length=1000,
                 subpackets_per_packet=1,
                 camera_dims=(128, 128, 2),
                 include_polarity=True,
                 allow_duplicate_events=False,
                 downscale=1):
        self.packets_fixed_by = packets_fixed_by
        if packets_fixed_by.startswith('dur'):
            # Convert milliseconds to microseconds
            self.subpacket_length = subpacket_length * 1000
        else:
            self.subpacket_length = subpacket_length
        self.subpackets_per_packet = subpackets_per_packet
        self.downscale = downscale
        if include_polarity:
            self.camera_dims = (camera_dims[0],) + (camera_dims[1],) + (2,)
        else:
            self.camera_dims = (camera_dims[0],) + (camera_dims[1],) + (1,)
        self.include_polarity = include_polarity
        self.sample_dims = (self.camera_dims[0] // downscale,
                            self.camera_dims[1] // downscale,
                            self.camera_dims[2] * subpackets_per_packet)
        self.allow_duplicate_events = allow_duplicate_events

        self.dataset = dataset
        if not inversion:
            # DVS gesture dataset is described and provided (and loaded) with
            # classes indexed from 1, not 0.
            # This modifier corrects for that offset
            do_y_inversion = False
            do_x_inversion = False
            print('DVS dataset: coordinates not inverted')
        else:
            do_y_inversion = True
            do_x_inversion = True
            # Coordinates delivered by matlab reading are inverted,
            # here we de-invert them
            print('DVS dataset: pyfile coordinates inverted')

        self.subject_list = []
        trial_durations = []
        label_offset = int(self.dataset[0]['label'])
        for trial in self.dataset:
            subj = int(trial['subject'])
            if subj not in self.subject_list:
                self.subject_list.append(subj)
            trial['label'] = int(trial['label']) - label_offset
            if do_y_inversion:
                # For camera dim of 128, calculation should be 127 - event
                # coord
                trial['events'][:, 1] = self.camera_dims[1] - trial[
                    'events'][:, 1] - 1
            if do_x_inversion:
                trial['events'][:, 0] = self.camera_dims[0] - trial[
                    'events'][:, 0] - 1
            trial_durations.append(trial['ev_times'][-1])
        self.trial_durations = trial_durations

    def get_random_samples(self,
                           class_list,
                           class_key=None,
                           samples_per_trial=5):
        """ Generates [samples_per_trial] samples from each subject/lighting
        condition/class.

        Args:
            class_list (list): classes selected for training
            class_key (list, optional): keys of selected classes. Defaults to
             None.
            samples_per_trial (int, optional): number of times that every sample
             have to be repeated. Defaults to 5.

        Returns:
            tuple: events and labels
        """
        if class_key is None:
            class_key = list(range(len(class_list)))
        outevents = []
        outlabels = []
        sample_windows = []
        for _, trial in enumerate(self.dataset):
            if int(trial['label']) in class_list:
                # Get latest start point (valid across all repeats)
                num_events = trial['ev_times'].shape[0]
                if self.packets_fixed_by.startswith('dur'):
                    latest_start_time = trial['ev_times'][-1] - (
                        self.subpacket_length * self.subpackets_per_packet)
                    if latest_start_time < 0:
                        print('ERROR: Sample durations requested is longer \
                            than the available data for this trial')
                        continue
                    latest_start_point = np.searchsorted(np.squeeze(
                        trial['ev_times']),
                                                         latest_start_time,
                                                         side='right') - 1
                elif self.packets_fixed_by.startswith('ev'):
                    latest_start_point = num_events - self.subpackets_per_packet * self.subpacket_length

                if latest_start_point < 0:
                    print("ERROR: Latest start point isn't valid")
                    continue

                # Build samples from multiple packets
                for _ in range(samples_per_trial):
                    sample_evs = np.zeros((0, 3), dtype=np.uint8)
                    start_point = np.random.randint(latest_start_point)
                    sample_start_point = start_point
                    sample_start_time = trial['ev_times'][start_point]
                    for packet in range(self.subpackets_per_packet):
                        # Set packet end point relative to start point
                        # Convention here is EXCLUSIVE end_point, for clarity
                        # in python
                        if self.packets_fixed_by.startswith('dur'):
                            end_time = sample_start_time + \
                                self.subpacket_length * (packet + 1)
                            end_point = np.searchsorted(np.squeeze(
                                trial['ev_times']),
                                                        end_time,
                                                        side='right')
                        elif self.packets_fixed_by.startswith('ev'):
                            end_point = start_point + self.subpacket_length
                        # Get the actual events
                        ev_array = np.copy(trial['events'][
                            np.squeeze(start_point):np.squeeze(end_point), :])
                        if self.downscale > 1:
                            ev_array[:, 0] = ev_array[:, 0] // self.downscale
                            ev_array[:, 1] = ev_array[:, 1] // self.downscale

                        # We use the events feature dimension to encode the
                        # temporal order of packets
                        if not self.include_polarity:
                            ev_array[:, 2] = 0
                        ev_array[:, 2] += self.camera_dims[2] * packet
                        sample_evs = np.concatenate((sample_evs, ev_array))

                        # Update start point for the next packet
                        start_point = end_point
                    outevents.append(sample_evs)
                    outlabels.append(class_key[class_list.index(
                        int(trial['label']))])
                    if self.packets_fixed_by.startswith('dur'):
                        sample_windows.append(end_point - sample_start_point)
                    elif self.packets_fixed_by.startswith('ev'):
                        sample_windows.append(trial['ev_times'][end_point - 1] -
                                              sample_start_time)

        if self.packets_fixed_by.startswith('dur'):
            mean_events = np.mean(np.array(sample_windows))
            # mean_duration is in msecs
            mean_duration = self.subpacket_length / 1000 * self.subpackets_per_packet
            # mean_ev_rate stores events per second
            mean_ev_rate = mean_events / mean_duration * 1000
        elif self.packets_fixed_by.startswith('ev'):
            mean_events = self.subpacket_length * self.subpackets_per_packet
            # ev_rate stores events per msecs
            ev_rates = mean_events / np.array(sample_windows) * 1000
            # mean_ev_rate stores events per second
            mean_ev_rate = np.mean(ev_rates) * 1000
        print('DVS event rate: ' + str(np.round(mean_ev_rate)) + ' events/sec')

        outlabels = np.array(outlabels)
        return outevents, outlabels

    def get_continuous_sample(self,
                              subj,
                              class_label,
                              class_key=None,
                              start_time=0,
                              stop_time=None):
        """ Generates series of temporally contiguous samples from a single
        subject+class.

        Args:
            subj (int): subject id
            class_label (int): label of the class
            class_key (int, optional): key of the class. Defaults to None.
            start_time (int, optional): samples starting point. Defaults to 0.
            stop_time (int, optional): samples ending point. Defaults to None.

        Returns:
            tuple: events, labels, plot events
        """
        if class_key is None:
            class_key = class_label
        outevents = []
        outlabels = []
        plotevents = []
        sample_windows = []
        for _, trial in enumerate(self.dataset):
            local_stop_time = stop_time
            if (int(trial['label']) == class_label) and (int(trial['subject'])
                                                         == subj):
                # Get latest start point (valid across all repeats)
                num_events = trial['ev_times'].shape[0]
                if local_stop_time is None:
                    local_stop_time = trial['ev_times'][-1]
                elif local_stop_time > trial['ev_times'][-1]:
                    print('ERROR: requested stop time later than the end of \
                         the trial')
                    print('Ending samples at ' + str(trial['ev_times'][-1]))
                    local_stop_time = trial['ev_times'][-1]
                if self.packets_fixed_by.startswith('dur'):
                    latest_start_time = local_stop_time - \
                        (self.subpacket_length * self.subpackets_per_packet)
                    if latest_start_time < 0:
                        print('ERROR: Sample durations requested is longer \
                            than the available data for this trial (or stop \
                            time is set too early)')
                        continue
                    latest_start_point = np.searchsorted(np.squeeze(
                        trial['ev_times']),
                                                         latest_start_time,
                                                         side='right') - 1
                elif self.packets_fixed_by.startswith('ev'):
                    if local_stop_time < trial['ev_times'][-1]:
                        stop_point = np.searchsorted(np.squeeze(
                            trial['ev_times']),
                                                     local_stop_time,
                                                     side='right') - 1
                    else:
                        stop_point = num_events
                    latest_start_point = stop_point - self.subpackets_per_packet * self.subpacket_length

                # Build samples from multiple packets
                if start_time > 0:
                    sample_start_point = np.searchsorted(np.squeeze(
                        trial['ev_times']),
                                                         start_time,
                                                         side='left')
                else:
                    sample_start_point = 0
                sample_start_time = start_time
                while sample_start_point < latest_start_point:
                    sample_evs = np.zeros((0, 3), dtype=np.uint8)
                    plot_evs = np.zeros((0, 3), dtype=np.int)
                    start_point = sample_start_point
                    for packet in range(self.subpackets_per_packet):
                        # Set packet end point relative to start point
                        # Convention here is EXCLUSIVE end_point, for clarity
                        # in python
                        if self.packets_fixed_by.startswith('dur'):
                            end_time = sample_start_time + \
                                self.subpacket_length * (packet + 1)
                            end_point = np.searchsorted(np.squeeze(
                                trial['ev_times']),
                                                        end_time,
                                                        side='right')
                        elif self.packets_fixed_by.startswith('ev'):
                            end_point = start_point + self.subpacket_length

                        # Get the actual events
                        plot_array = np.copy(trial['events'][
                            np.squeeze(start_point):np.squeeze(end_point), :])
                        plot_evs = np.concatenate((plot_evs, plot_array))

                        ev_array = np.copy(trial['events'][
                            np.squeeze(start_point):np.squeeze(end_point), :])
                        if self.downscale > 1:
                            ev_array[:, 0] = ev_array[:, 0] // self.downscale
                            ev_array[:, 1] = ev_array[:, 1] // self.downscale
                        if not self.include_polarity:
                            ev_array[:, 2] = 0
                        ev_array[:, 2] += self.camera_dims[2] * packet
                        sample_evs = np.concatenate((sample_evs, ev_array))

                        # Update start point for the next packet
                        start_point = end_point
                    outevents.append(sample_evs)
                    outlabels.append(class_key)
                    plotevents.append(plot_evs)
                    if self.packets_fixed_by.startswith('dur'):
                        sample_windows.append(end_point - sample_start_point)
                        sample_start_time += self.subpacket_length
                        sample_start_point = np.searchsorted(np.squeeze(
                            trial['ev_times']),
                                                             sample_start_time,
                                                             side='right')
                    elif self.packets_fixed_by.startswith('ev'):
                        sample_windows.append(trial['ev_times'][end_point - 1] -
                                              sample_start_time)
                        sample_start_point += self.subpacket_length
        outlabels = np.array(outlabels)
        return outevents, outlabels, plotevents

    def events2matrix(self, event_packet):
        """ Transforms a DVS aer event into a numpy array.

        Args:
            event_packet (numpy.ndarray): input events packet

        Returns:
            numpy.ndarray: matrix of events
        """
        if self.allow_duplicate_events:
            out_mat = np.zeros((self.sample_dims))
            for ind in range(event_packet.shape[0]):
                out_mat[event_packet[ind, 1], event_packet[ind, 0],
                        event_packet[ind, 2]] += 1
        else:
            length = np.prod(np.array(self.sample_dims, dtype=int))
            out_mat = np.zeros((length, 1))
            indices = np.ravel_multi_index(
                (event_packet[:, 1], event_packet[:, 0], event_packet[:, 2]),
                (self.sample_dims))
            out_mat[indices] = 1
            out_mat = np.reshape(out_mat, (self.sample_dims))
        return out_mat.astype(np.float16)

    def samples_evs2images(self, samples_events):
        """ Transforms a list of DVS aer events into a matrix suitable for
        TensorFlow training.

        Args:
            samples_events (list): list of DVS events

        Returns:
            np.ndarray: events as an image array
        """
        outmat = []
        for sample in samples_events:
            sample_mat = self.events2matrix(sample)
            outmat.append(sample_mat)
        return np.array(outmat)

    def augment_data(self,
                     samples_events,
                     horizontal_flip=False,
                     vertical_flip=False,
                     polarity_flip=False,
                     rotation_range=0,
                     width_shift_range=1.0,
                     height_shift_range=1.0):
        """ Performs data augmentations on sample events.

        Args:
            samples_events (list): list of DVS events to process
            horizontal_flip (bool, optional): enable horizontal flip. Defaults
             to False.
            vertical_flip (bool, optional): enable vertical flip. Defaults to
             False.
            polarity_flip (bool, optional): enable polarity flip. Defaults to
             False.
            rotation_range (int, optional): rotation angle range in degrees.
             Defaults to 0.
            width_shift_range (float, optional): range of width shift. Defaults
             to 1.0.
            height_shift_range (float, optional): range of height shift.
             Defaults to 1.0.

        Returns:
            list: augmented events
        """
        outevents = []
        for sample in samples_events:
            samp = np.copy(sample).astype(np.float)
            # To make subsequent transforms easier, centre matrix around (0, 0)
            samp[:, 0] -= self.sample_dims[1] // 2
            samp[:, 1] -= self.sample_dims[0] // 2

            if horizontal_flip and random.random() > 0.5:
                samp[:, 0] = -samp[:, 0]
            if vertical_flip and random.random() > 0.5:
                samp[:, 1] = -samp[:, 1]
            if polarity_flip and random.random() > 0.5:
                # shift all odd numbers by -2
                samp[(samp[:, 2] % 2).astype(np.bool), 2] -= 2
                samp[:, 2] += 1
            if rotation_range > 0:
                # degrees to radians
                theta = random.uniform(-rotation_range,
                                       rotation_range) * np.pi / 180
                rot_matrix = np.array([[np.cos(theta), -np.sin(theta)],
                                       [np.sin(theta),
                                        np.cos(theta)]])
                samp[:, :2] = np.dot(rot_matrix, samp[:, :2].T).T
            if width_shift_range != 1.0:
                scale = random.uniform(1, width_shift_range)
                if random.random() > 0.5:
                    # Scale Up
                    samp[:, 0] = samp[:, 0] * scale
                else:
                    # Scale down
                    samp[:, 0] = samp[:, 0] / scale
            if height_shift_range != 1.0:
                scale = random.uniform(1, width_shift_range)
                if random.random() > 0.5:
                    # Scale Up
                    samp[:, 1] = samp[:, 1] * scale
                else:
                    # Scale down
                    samp[:, 1] = samp[:, 1] / scale
            samp[:, 0] += self.sample_dims[1] // 2
            samp[:, 1] += self.sample_dims[0] // 2
            samp = np.round(samp)
            out_of_range = np.where(
                np.logical_or(
                    np.logical_or(samp[:, 0] >= self.sample_dims[1],
                                  samp[:, 0] < 0),
                    np.logical_or(samp[:, 1] >= self.sample_dims[0],
                                  samp[:, 1] < 0)))
            samp = np.delete(samp, out_of_range, axis=0)
            outevents.append(samp.astype(np.uint8))
        return outevents



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Parameters scheduling policy module.

Originated from: https://github.com/fastai/fastai/blob/master/nbs/14_callback.schedule.ipynb
"""

import math

import numpy as np

from keras.callbacks import Callback


class _Annealer:
    # pylint: disable=too-few-public-methods
    def __init__(self, f, start, end):
        self.f = f
        self.start = start
        self.end = end

    def __call__(self, pos):
        return self.f(self.start, self.end, pos)


def SchedLin(start, end):
    """ Linear schedule function from `start` to `end`.

    Args:
        start (float): starting point
        end (float): ending point

    Returns:
        function: the scheduling function
    """

    def sched_lin(start, end, pos):
        return start + pos * (end - start)

    return _Annealer(sched_lin, start, end)


def SchedCos(start, end):
    """ Cosine schedule function from `start` to `end`.

    Args:
        start (float): starting point
        end (float): ending point

    Returns:
        function: the scheduling function
    """

    def sched_cos(start, end, pos):
        return start + (1 + math.cos(math.pi * (1 - pos))) * (end - start) / 2

    return _Annealer(sched_cos, start, end)


def SchedNo(start, end):
    """ Constant schedule function with `start` value.

    Args:
        start (float): starting point
        end (float): ending point

    Returns:
        function: the scheduling function
    """

    def sched_no(start, end, pos):  # pylint: disable=unused-argument
        return start

    return _Annealer(sched_no, start, end)


def SchedExp(start, end):
    """ Exponential schedule function from `start` to `end`.

    Args:
        start (float): starting point
        end (float): ending point

    Returns:
        function: the scheduling function
    """
    if start == 0:
        raise ValueError("Starting point cannot be 0 for SchedExp.")

    def sched_exp(start, end, pos):
        return start * (end / start)**pos

    return _Annealer(sched_exp, start, end)


def SchedPoly(start, end, power):
    """ Polynomial schedule (of `power`) function from `start` to `end`.

    Args:
        start (float): starting point
        end (float): ending point
        power (float): power of the polynomial

    Returns:
        function: the scheduling function
    """

    def _inner(pos):
        return start + (end - start) * pos**power

    return _inner


def combine_scheds(pcts, scheds):
    """ Combine `scheds` according to `pcts` in one function.

    The generated function will use scheds[0] from 0 to pcts[0]
    then scheds[1] from pcts[0] to pcts[0]+pcts[1] and so forth.

    example:
        all_pos = np.linspace(0., 1, 100)
        pcts = np.array([0.3, 0.7])
        scheds = [SchedCos(0.3, 0.6), SchedCos(0.6, 0.2)]
        sched_comb = combine_scheds(pcts, scheds)
        all_lrs = []
        for pos in all_pos:
            all_lrs.append(sched_comb(pos))
        plt.plot(all_pos, all_lrs)
        plt.show()

    example 2 :
        p = np.linspace(0., 1, 100)
        f = combine_scheds([0.3, 0.2, 0.5], [SchedLin(0., 1.), SchedNo(1., 1.),
                SchedCos(1., 0.)])
        plt.plot(p, [f(o) for o in p])
        plt.show()

    Args:
        pcts (list): list of positive numbers that add up to 1 and is the same
            length as scheds.
        scheds (list): list of scheduling functions

    Returns:
        function: the combined scheduling function
    """
    assert sum(pcts) == 1.
    pcts = np.hstack((np.array(0,), pcts))

    assert np.all(pcts >= 0)
    pcts = np.cumsum(pcts, 0)

    def _inner(pos):
        if int(pos) == 1:
            return scheds[-1](1.)

        idx = np.max(np.where(pos >= pcts))
        actual_pos = (pos - pcts[idx]) / (pcts[idx + 1] - pcts[idx])
        return scheds[idx](actual_pos)

    return _inner


class ParamScheduler(Callback):
    """ Schedule hyper-parameters according to `scheds`.

    # example for linear scheduler for LR
        sched = {'lr': SchedLin(1e-3, 1e-2)}

    Args:
        scheds (dict): dictionary with one key for each hyper-parameter you want
            to schedule, with either a scheduler or a list of schedulers as
            values (in the second case, the list must have the same length as
            the number of parameters groups of the optimizer).
        total_num_iterations (int): number of iterations
    """

    def __init__(self, scheds, total_num_iterations):
        super().__init__()

        self.scheds = scheds
        # pct_train: from 0. to 1., the percentage of training iterations
        # completed
        self.pct_train = 0  # is rewritten at each batch
        self.history = {}

        if total_num_iterations <= 0:
            raise ValueError(f"total_num_iterations must be > 0 but received \
                {total_num_iterations}")

        self.total_num_iterations = total_num_iterations  # static
        self.cum_batch_num = 0  # keep adding 1 to this at every batch

    def _update_val(self, pct):
        # pct is the percent training done
        for n, f in self.scheds.items():
            #pylint: disable=protected-access
            self.model.optimizer._set_hyper(n, f(pct))

    def on_batch_begin(self, batch, logs=None):
        # before_batch: set the proper hyper-parameters in the optimizer
        self._update_val(self.pct_train)
        # update the tracking info
        self.cum_batch_num += 1
        self.pct_train = self.cum_batch_num / self.total_num_iterations

    def on_batch_end(self, batch, logs=None):
        # after_batch: record hyper-parameters of this batch
        logs = logs or {}
        # keep in memory the current values
        for p in self.scheds.keys():
            self.history.setdefault(p, []).append(
                self.model.optimizer.get_config()[p])
        # logs on batch end is a dict with aggregated metric results up until
        # this batch.
        for k, v in logs.items():
            self.history.setdefault(k, []).append(v)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
DS-CNN model definition for CIFAR10 classification.
"""

from keras.layers import Input, Rescaling
from keras import Model
from keras.utils.data_utils import get_file

from cnn2snn import load_quantized_model, quantize

from ..layer_blocks import conv_block, separable_conv_block

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/ds_cnn/'


def ds_cnn_cifar10(input_shape=(32, 32, 3),
                   classes=10,
                   weight_quantization=0,
                   activ_quantization=0,
                   input_weight_quantization=None,
                   input_scaling=(255, 0)):
    """Instantiates a MobileNet-like model for the "CIFAR-10" example.
    This model is based on the MobileNet architecture, mainly with fewer layers.
    The weights and activations are quantized such that it can be converted into
    an Akida model.

    This architecture is originated from https://arxiv.org/abs/1704.04861 and
    inspired from https://arxiv.org/pdf/1711.07128.pdf.

    Args:
        input_shape (tuple): input shape tuple of the model
        classes (int): number of classes to classify images into
        weight_quantization (int): sets all weights in the model to have
            a particular quantization bitwidth except for the weights in the
            first layer.

            * '0' implements floating point 32-bit weights
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization (int): sets all activations in the model to have a.
            particular activation quantization bitwidth.

            * '0' implements floating point 32-bit activations.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_weight_quantization (int): sets weight quantization in the first
            layer. Defaults to weight_quantization value.

            * 'None' implements the same bitwidth as the other weights.
            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_scaling (tuple, optional): scale factor and offset to apply to
            inputs. Defaults to (255, 0). Note that following Akida convention,
            the scale factor is an integer used as a divider.

    Returns:
        keras.Model: a Keras model for DS-CNN/CIFAR-10
    """

    # Overrides input weight quantization if None
    if input_weight_quantization is None:
        input_weight_quantization = weight_quantization

    img_input = Input(shape=input_shape)

    if input_scaling is None:
        x = img_input
    else:
        scale, offset = input_scaling
        x = Rescaling(1. / scale, offset)(img_input)

    x = conv_block(x,
                   filters=128,
                   name='conv_0',
                   kernel_size=(3, 3),
                   padding='same',
                   use_bias=False,
                   add_batchnorm=True,
                   add_activation=True)

    x = separable_conv_block(x,
                             filters=128,
                             kernel_size=(3, 3),
                             name='separable_1',
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True)

    x = separable_conv_block(x,
                             filters=256,
                             kernel_size=(3, 3),
                             name='separable_2',
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True)

    x = separable_conv_block(x,
                             filters=256,
                             kernel_size=(3, 3),
                             name='separable_3',
                             padding='same',
                             use_bias=False,
                             pooling='max',
                             add_batchnorm=True,
                             add_activation=True)

    x = separable_conv_block(x,
                             filters=512,
                             kernel_size=(3, 3),
                             name='separable_4',
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True)

    x = separable_conv_block(x,
                             filters=512,
                             kernel_size=(3, 3),
                             name='separable_5',
                             padding='same',
                             use_bias=False,
                             pooling='max',
                             add_batchnorm=True,
                             add_activation=True)

    x = separable_conv_block(x,
                             filters=512,
                             kernel_size=(3, 3),
                             name='separable_6',
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True)

    x = separable_conv_block(x,
                             filters=512,
                             kernel_size=(3, 3),
                             name='separable_7',
                             padding='same',
                             use_bias=False,
                             pooling='max',
                             add_batchnorm=True,
                             add_activation=True)

    x = separable_conv_block(x,
                             filters=1024,
                             kernel_size=(3, 3),
                             name='separable_8',
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True)

    x = separable_conv_block(x,
                             filters=1024,
                             kernel_size=(3, 3),
                             name='separable_9',
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True)

    x = separable_conv_block(x,
                             filters=classes,
                             kernel_size=(3, 3),
                             name='separable_10',
                             padding='same',
                             use_bias=False,
                             pooling='global_avg',
                             add_batchnorm=False,
                             add_activation=False)

    model = Model(img_input, x, name='ds_cnn_cifar10')

    if ((weight_quantization != 0) or (activ_quantization != 0) or
        (input_weight_quantization != 0)):
        return quantize(model, weight_quantization, activ_quantization,
                        input_weight_quantization)
    return model


def ds_cnn_cifar10_pretrained():
    """
    Helper method to retrieve a `ds_cnn_cifar10` model that was trained on
    CIFAR10 dataset.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'ds_cnn_cifar10_iq4_wq4_aq4.h5'
    file_hash = '954f61fb6014a41a198cfa313fae5c40273614033da16b3f0f6b9a5db0ac642c'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)



from os import path
from setuptools import setup, find_packages

# Read the contents of the README file
directory = path.abspath(path.dirname(__file__))
with open(path.join(directory, 'README'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='akida_models',
      version='1.1.0',
      description='Akida Models',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='David Corvoysier',
      author_email='dcorvoysier@brainchip.com',
      url='https://doc.brainchipinc.com',
      license='Apache 2.0',
      license_files=['LICENSE', 'LICENSE.3rdparty'],
      packages=find_packages(),
      entry_points={
        'console_scripts': [
            'akida_models = akida_models.cli:main',
            'cifar10_train = akida_models.cifar10.cifar10_train:main',
            'utk_face_train = akida_models.utk_face.utk_face_train:main',
            'kws_train = akida_models.kws.kws_train:main',
            'modelnet40_train = akida_models.modelnet40.modelnet40_train:main',
            'yolo_train = akida_models.detection.yolo_train:main',
            'dvs_train = akida_models.dvs.dvs_train:main',
            'cwru_train = akida_models.cwru.cwru_train:main',
            'mnist_train = akida_models.mnist.mnist_train:main'
        ]
      },
      install_requires=['cnn2snn>=2.1.1', 'scipy', 'opencv-python', 'mtcnn',
        'imgaug', 'dv', 'trimesh'],
      python_requires='>=3.7')



#!/usr/bin/env python
# ******************************************************************************
# MIT License
#
# Copyright (c) 2017 Ngoc Anh Huynh
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ******************************************************************************
"""
Data generator for YOLO training.
"""

import copy
import cv2
import numpy as np

from imgaug import augmenters as iaa
from keras.utils.data_utils import Sequence

from .processing import BoundingBox


class BatchGenerator(Sequence):
    """ Data generator used for the training process.
    """

    def __init__(self,
                 input_shape,
                 data,
                 grid_size,
                 labels,
                 anchors,
                 batch_size,
                 shuffle=True,
                 jitter=True):
        self._input_shape = input_shape
        self._data = data
        self._data_length = len(data)
        self._grid_size = grid_size
        self._labels = labels
        self._batch_size = batch_size

        self._shuffle = shuffle
        self._jitter = jitter

        self._anchors = [
            BoundingBox(0, 0, anchors[i][0], anchors[i][1])
            for i in range(len(anchors))
        ]

        # augmentors by https://github.com/aleju/imgaug
        sometimes = lambda aug: iaa.Sometimes(0.5, aug)

        # Define a sequence of augmentation steps that will be applied to every
        # image. All augmenters with per_channel=0.5 will sample one value per
        # image in 50% of all cases. In all other cases they will sample new
        # values per channel.
        self._aug_pipe = iaa.Sequential(
            [
                # apply the following augmenters to most images
                sometimes(iaa.Affine()),
                # execute 0 to 5 of the following (less important) augmenters
                # per image. Don't execute all of them, as that would often be
                # way too strong
                iaa.SomeOf(
                    (0, 5),
                    [
                        iaa.OneOf([
                            # blur images with a sigma between 0 and 3.0
                            iaa.GaussianBlur((0, 3.0)),
                            # blur image using local means (kernel sizes between
                            # 2 and 7)
                            iaa.AverageBlur(k=(2, 7)),
                            # blur image using local medians (kernel sizes
                            # between 3 and 11)
                            iaa.MedianBlur(k=(3, 11)),
                        ]),
                        # sharpen images
                        iaa.Sharpen(alpha=(0, 1.0), lightness=(0.75, 1.5)),
                        # add gaussian noise
                        iaa.AdditiveGaussianNoise(
                            loc=0, scale=(0.0, 0.05 * 255), per_channel=0.5),
                        # randomly remove up to 10% of the pixels
                        iaa.OneOf([
                            iaa.Dropout((0.01, 0.1), per_channel=0.5),
                        ]),
                        # change brightness of images
                        iaa.Add((-10, 10), per_channel=0.5),
                        # change brightness of images
                        iaa.Multiply((0.5, 1.5), per_channel=0.5),
                        # improve or worsen the contrast
                        iaa.LinearContrast((0.5, 2.0), per_channel=0.5),
                    ],
                    random_order=True)
            ],
            random_order=True)

        if shuffle:
            np.random.shuffle(self._data)

    def __len__(self):
        return int(np.ceil(float(self._data_length) / self._batch_size))

    def __getitem__(self, idx):
        lower_bound = idx * self._batch_size
        upper_bound = (idx + 1) * self._batch_size

        if upper_bound > self._data_length:
            upper_bound = self._data_length
            lower_bound = upper_bound - self._batch_size

        instance_count = 0

        x_batch = np.zeros((upper_bound - lower_bound, self._input_shape[1],
                            self._input_shape[0], self._input_shape[2]))

        y_batch = np.zeros(
            (upper_bound - lower_bound, self._grid_size[1], self._grid_size[0],
             len(self._anchors),
             4 + 1 + len(self._labels)))  # desired network output

        for train_instance in self._data[lower_bound:upper_bound]:
            # augment input image and fix object's position and size
            img, all_objs = self.aug_image(train_instance, jitter=self._jitter)

            for obj in all_objs:
                if obj['x2'] > obj['x1'] and obj['y2'] > obj['y1'] and obj[
                        'label'] in self._labels:
                    center_x = .5 * (obj['x1'] + obj['x2'])
                    center_x = center_x / (float(self._input_shape[0]) /
                                           self._grid_size[0])
                    center_y = .5 * (obj['y1'] + obj['y2'])
                    center_y = center_y / (float(self._input_shape[1]) /
                                           self._grid_size[0])

                    grid_x = int(np.floor(center_x))
                    grid_y = int(np.floor(center_y))

                    if grid_x < self._grid_size[0] and grid_y < self._grid_size[
                            1]:
                        obj_indx = self._labels.index(obj['label'])

                        center_w = (obj['x2'] - obj['x1']) / (
                            float(self._input_shape[0]) / self._grid_size[0])
                        center_h = (obj['y2'] - obj['y1']) / (
                            float(self._input_shape[1]) / self._grid_size[0])

                        box = [center_x, center_y, center_w, center_h]

                        # find the anchor that best predicts this box
                        best_anchor = -1
                        max_iou = -1

                        shifted_box = BoundingBox(0, 0, center_w, center_h)

                        for anchor_id, anchor in enumerate(self._anchors):
                            iou = shifted_box.iou(anchor)

                            if max_iou < iou:
                                best_anchor = anchor_id
                                max_iou = iou

                        # assign ground truth x, y, w, h, confidence and class
                        # probs to y_batch
                        y_batch[instance_count, grid_y, grid_x, best_anchor,
                                0:4] = box
                        y_batch[instance_count, grid_y, grid_x, best_anchor,
                                4] = 1.
                        y_batch[instance_count, grid_y, grid_x, best_anchor,
                                5 + obj_indx] = 1

            # assign input image to x_batch
            x_batch[instance_count] = img

            # increase instance counter in current batch
            instance_count += 1

        return x_batch, y_batch

    def on_epoch_end(self):
        if self._shuffle:
            np.random.shuffle(self._data)

    def aug_image(self, train_instance, jitter):
        """ Performs data augmentation on an image.

        Args:
            train_instance (dict): the dictionnary of input data (image and
             boxes)
            jitter (bool): enable jitter in the augmentation pipeline

        Raises:
            ValueError: if the number of channel in the input shape is wrong
             (ie. different from 1 or 3).
            RuntimeError: if the image cannot be found

        Returns:
            tuple: the augmented image and the associated boxes
        """
        image_path = train_instance['image_path']
        if self._input_shape[2] == 1:
            image = cv2.imread(image_path, cv2.IMREAD_GRAYSCALE)
        elif self._input_shape[2] == 3:
            image = cv2.imread(image_path)
        else:
            raise ValueError("Invalid number of image channels.")

        if image is None:
            raise RuntimeError(f'Cannot find {image_path}')

        h = image.shape[0]
        w = image.shape[1]
        all_objs = copy.deepcopy(train_instance['boxes'])

        if jitter:
            # scale the image
            scale = np.random.uniform() / 10. + 1.
            image = cv2.resize(image, (0, 0), fx=scale, fy=scale)

            # translate the image
            max_offx = (scale - 1.) * w
            max_offy = (scale - 1.) * h
            offx = int(np.random.uniform() * max_offx)
            offy = int(np.random.uniform() * max_offy)

            image = image[offy:(offy + h), offx:(offx + w)]

            # flip the image
            flip = np.random.binomial(1, .5)
            if flip > 0.5:
                image = cv2.flip(image, 1)

            image = self._aug_pipe.augment_image(image)

        # resize the image to standard size
        image = cv2.resize(image, (self._input_shape[0], self._input_shape[1]))
        if self._input_shape[2] == 1:
            image = image[..., np.newaxis]
        image = image[..., ::-1]

        # fix object's position and size
        for obj in all_objs:
            for attr in ['x1', 'x2']:
                if jitter:
                    obj[attr] = int(obj[attr] * scale - offx)

                obj[attr] = int(obj[attr] * float(self._input_shape[0]) / w)
                obj[attr] = max(min(obj[attr], self._input_shape[0]), 0)

            for attr in ['y1', 'y2']:
                if jitter:
                    obj[attr] = int(obj[attr] * scale - offy)

                obj[attr] = int(obj[attr] * float(self._input_shape[1]) / h)
                obj[attr] = max(min(obj[attr], self._input_shape[1]), 0)

            if jitter and flip > 0.5:
                xmin = obj['x1']
                obj['x1'] = self._input_shape[0] - obj['x2']
                obj['x2'] = self._input_shape[0] - xmin
        return image, all_objs



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
VGG model definition for UTKFace regression.
"""

from keras import Model
from keras.layers import Dropout, Flatten, Input, Rescaling
from keras.utils.data_utils import get_file

from cnn2snn import load_quantized_model, quantize

from ..layer_blocks import conv_block, dense_block

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/vgg/'


def vgg_utk_face(input_shape=(32, 32, 3),
                 weight_quantization=0,
                 activ_quantization=0,
                 input_weight_quantization=None,
                 input_scaling=(127, -1)):
    """Instantiates a VGG-like model for the regression example on age
    estimation using UTKFace dataset.

    Args:
        input_shape (tuple): input shape tuple of the model
        weight_quantization (int): sets all weights in the model to have
            a particular quantization bitwidth except for the weights in the
            first layer.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization (int): sets all activations in the model to have a
            particular activation quantization bitwidth.

            * '0' implements floating point 32-bit activations.
            * '1' through '8' implements n-bit weights where n is from 1-8 bits.
        input_weight_quantization (int): sets weight quantization in the first
            layer. Defaults to weight_quantization value.

            * 'None' implements the same bitwidth as the other weights.
            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_scaling (tuple, optional): scale factor and offset to apply to
            inputs. Defaults to (127, -1). Note that following Akida convention,
            the scale factor is an integer used as a divider.

    Returns:
        keras.Model: a Keras model for VGG/UTKFace
    """
    # Overrides input weight quantization if None
    if input_weight_quantization is None:
        input_weight_quantization = weight_quantization

    img_input = Input(shape=input_shape)

    if input_scaling is None:
        x = img_input
    else:
        scale, offset = input_scaling
        x = Rescaling(1. / scale, offset)(img_input)

    x = conv_block(x,
                   filters=32,
                   kernel_size=(3, 3),
                   name='conv_0',
                   use_bias=False,
                   add_batchnorm=True,
                   add_activation=True)

    x = conv_block(x,
                   filters=32,
                   kernel_size=(3, 3),
                   name='conv_1',
                   padding='same',
                   pooling='max',
                   pool_size=2,
                   use_bias=False,
                   add_batchnorm=True,
                   add_activation=True)

    x = Dropout(0.3)(x)

    x = conv_block(x,
                   filters=64,
                   kernel_size=(3, 3),
                   padding='same',
                   name='conv_2',
                   use_bias=False,
                   add_batchnorm=True,
                   add_activation=True)

    x = conv_block(x,
                   filters=64,
                   kernel_size=(3, 3),
                   padding='same',
                   name='conv_3',
                   pooling='max',
                   pool_size=2,
                   use_bias=False,
                   add_batchnorm=True,
                   add_activation=True)

    x = Dropout(0.3)(x)

    x = conv_block(x,
                   filters=84,
                   kernel_size=(3, 3),
                   padding='same',
                   name='conv_4',
                   use_bias=False,
                   add_batchnorm=True,
                   add_activation=True)

    x = Dropout(0.3)(x)
    x = Flatten()(x)

    x = dense_block(x,
                    units=64,
                    name='dense_1',
                    use_bias=False,
                    add_batchnorm=True,
                    add_activation=True)

    x = dense_block(x, units=1, name='dense_2', add_activation=False)

    model = Model(img_input, x, name='vgg_utk_face')

    if ((weight_quantization != 0) or (activ_quantization != 0) or
        (input_weight_quantization != 0)):
        return quantize(model, weight_quantization, activ_quantization,
                        input_weight_quantization)
    return model


def vgg_utk_face_pretrained():
    """
    Helper method to retrieve a `vgg_utk_face` model that was trained on
    UTK Face dataset.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'vgg_utk_face_iq8_wq2_aq2.h5'
    file_hash = 'e688686ad311e36f14e49c720e85461620b7406329cef00fa7afe79b33080b01'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
VGG model definition for ImageNet classification.
"""

from keras import Model, regularizers
from keras.layers import (Input, Activation, Dropout, Flatten, Rescaling)
from keras.utils.data_utils import get_file

from cnn2snn import quantize, load_quantized_model

from ..layer_blocks import conv_block, dense_block

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/vgg/'


def vgg_imagenet(input_shape=(224, 224, 3),
                 classes=1000,
                 include_top=True,
                 pooling=None,
                 weight_quantization=0,
                 activ_quantization=0,
                 input_weight_quantization=None,
                 input_scaling=(128, -1)):
    """Instantiates a VGG11 architecture with reduced number of filters in
    convolutional layers (i.e. a quarter of the filters of the original
    implementation of https://arxiv.org/pdf/1409.1556.pdf).

    Args:
        input_shape (tuple, optional): input shape tuple. Defaults to (224, 224,
            3).
        classes (int, optional): optional number of classes to classify images
            into. Defaults to 1000.
        include_top (bool, optional): whether to include the classification
            layers at the top of the model. Defaults to True.
        pooling (str, optional): Optional pooling mode for feature extraction
            when `include_top` is `False`. Defaults to None.

            * `None` means that the output of the model will be the 4D tensor
              output of the last convolutional block.
            * `avg` means that global average pooling will be applied to the
              output of the last convolutional block, and thus the output of the
              model will be a 2D tensor.
        weight_quantization (int, optional): sets all weights in the model to
            have a particular quantization bitwidth except for the weights in
            the first layer. Defaults to 0.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization (int, optional): sets all activations in the model to
            have a particular activation quantization bitwidth. Defaults to 0.

            * '0' implements floating point 32-bit activations.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_weight_quantization(int, optional): sets weight quantization in
            the first layer. Defaults to weight_quantization value.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_scaling (tuple, optional): scale factor and offset to apply to
            inputs. Defaults to (128, -1). Note that following Akida convention,
            the scale factor is an integer used as a divider.

    Returns:
        keras.Model: a Keras model for VGG/ImageNet

    """
    # check if overrides have been provided and override
    if input_weight_quantization is None:
        input_weight_quantization = weight_quantization

    # Define weight regularization
    weight_regularizer = regularizers.l2(4e-5)

    img_input = Input(shape=input_shape)

    if input_scaling is None:
        x = img_input
    else:
        scale, offset = input_scaling
        x = Rescaling(1. / scale, offset)(img_input)

    # Block 1
    x = conv_block(x,
                   filters=16,
                   name='block1_conv1',
                   kernel_size=(3, 3),
                   padding='same',
                   kernel_regularizer=weight_regularizer,
                   add_batchnorm=True,
                   add_activation=True,
                   pooling='max',
                   pool_size=(2, 2))

    # Block 2
    x = conv_block(x,
                   filters=32,
                   name='block2_conv1',
                   kernel_size=(3, 3),
                   padding='same',
                   kernel_regularizer=weight_regularizer,
                   add_batchnorm=True,
                   add_activation=True,
                   pooling='max',
                   pool_size=(2, 2))

    # Block 3
    x = conv_block(x,
                   filters=64,
                   name='block3_conv1',
                   kernel_size=(3, 3),
                   padding='same',
                   kernel_regularizer=weight_regularizer,
                   add_batchnorm=True,
                   add_activation=True)

    x = conv_block(x,
                   filters=64,
                   name='block3_conv2',
                   kernel_size=(3, 3),
                   padding='same',
                   kernel_regularizer=weight_regularizer,
                   add_batchnorm=True,
                   add_activation=True,
                   pooling='max',
                   pool_size=(2, 2))

    # Block 4
    x = conv_block(x,
                   filters=128,
                   name='block4_conv1',
                   kernel_size=(3, 3),
                   padding='same',
                   kernel_regularizer=weight_regularizer,
                   add_batchnorm=True,
                   add_activation=True)

    x = conv_block(x,
                   filters=128,
                   name='block4_conv2',
                   kernel_size=(3, 3),
                   padding='same',
                   kernel_regularizer=weight_regularizer,
                   add_batchnorm=True,
                   add_activation=True,
                   pooling='max',
                   pool_size=(2, 2))

    # Block 5
    x = conv_block(x,
                   filters=128,
                   name='block5_conv1',
                   kernel_size=(3, 3),
                   padding='same',
                   kernel_regularizer=weight_regularizer,
                   add_batchnorm=True,
                   add_activation=True)

    layer_pooling = 'global_avg' if pooling == 'avg' else 'max'
    x = conv_block(x,
                   filters=128,
                   name='block5_conv2',
                   kernel_size=(3, 3),
                   padding='same',
                   kernel_regularizer=weight_regularizer,
                   add_batchnorm=True,
                   add_activation=True,
                   pooling=layer_pooling,
                   pool_size=(2, 2))

    if include_top:
        # Classification block
        x = Flatten(name='flatten')(x)
        x = dense_block(x,
                        units=4096,
                        name='fc1',
                        add_batchnorm=True,
                        add_activation=True)
        x = Dropout(0.5, name='dropout_1')(x)
        x = dense_block(x,
                        units=4096,
                        name='fc2',
                        add_batchnorm=True,
                        add_activation=True)
        x = Dropout(0.5, name='dropout_2')(x)
        x = dense_block(x,
                        units=classes,
                        name='predictions',
                        add_batchnorm=False,
                        add_activation=False)

        act_function = 'softmax' if classes > 1 else 'sigmoid'
        x = Activation(act_function, name=f'act_{act_function}')(x)

    # Create model
    model = Model(img_input, x, name='vgg11_%s_%s' % (input_shape[0], classes))

    if ((weight_quantization != 0) or (activ_quantization != 0) or
        (input_weight_quantization != 0)):
        return quantize(model, weight_quantization, activ_quantization,
                        input_weight_quantization)
    return model


def vgg_imagenet_pretrained():
    """
    Helper method to retrieve a `vgg_imagenet` model that was trained on
    ImageNet dataset.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'vgg11_imagenet_224_iq8_wq4_aq4.h5'
    file_hash = '6e806caacaa82d66f8cfc68de04d68d36ecca12570580fa61e770b14548b857b'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)



"""
Imports models.
"""

from .dvs.model_convtiny_gesture import (convtiny_dvs_gesture,
                                         convtiny_gesture_pretrained)
from .dvs.model_convtiny_handy import (convtiny_dvs_handy,
                                       convtiny_handy_samsung_pretrained)
from .cifar10.model_ds_cnn import ds_cnn_cifar10, ds_cnn_cifar10_pretrained
from .cifar10.model_vgg import vgg_cifar10, vgg_cifar10_pretrained
from .imagenet.model_mobilenet import (mobilenet_imagenet,
                                       mobilenet_imagenet_pretrained)
from .imagenet.model_vgg import (vgg_imagenet, vgg_imagenet_pretrained)
from .imagenet.model_mobilenet_edge import (mobilenet_edge_imagenet,
                                            mobilenet_edge_imagenet_pretrained)
from .imagenet.model_akidanet import (
    akidanet_imagenet, akidanet_imagenet_pretrained,
    akidanet_cats_vs_dogs_pretrained, akidanet_imagenette_pretrained,
    akidanet_faceidentification_pretrained,
    akidanet_faceverification_pretrained, akidanet_melanoma_pretrained,
    akidanet_odir5k_pretrained, akidanet_retinal_oct_pretrained,
    akidanet_ecg_pretrained, akidanet_plantvillage_pretrained)
from .imagenet.model_akidanet_edge import (
    akidanet_edge_imagenet, akidanet_edge_imagenet_pretrained,
    akidanet_faceidentification_edge_pretrained)
from .kws.model_ds_cnn import ds_cnn_kws, ds_cnn_kws_pretrained
from .modelnet40.model_pointnet_plus import (pointnet_plus_modelnet40,
                                             pointnet_plus_modelnet40_pretrained
                                            )
from .utk_face.model_vgg import vgg_utk_face, vgg_utk_face_pretrained
from .cse2018.model_tse import tse_mlp_cse2018, tse_mlp_cse2018_pretrained
from .tabular_data import tabular_data
from .detection.model_yolo import (yolo_base, yolo_widerface_pretrained,
                                   yolo_voc_pretrained)
from .cwru.model_convtiny import convtiny_cwru, convtiny_cwru_pretrained
from .gamma_constraint import add_gamma_constraint
from .filter_pruning import delete_filters, prune_model

from .mnist.model_gxnor import gxnor_mnist, gxnor_mnist_pretrained



# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Utils for PointNet.
"""

import tensorflow as tf


def shift_to_7bit(values, val_min, val_max):
    """ Normalize data to 7 bits
    Args:
        values (tf.Tensor): input data to normalize.
        val_min (int): minimum input data value.
        val_max (int): maximum input data value.
    Returns:
        tf.Tensor: data normalized to 7 bits.
    """
    return tf.clip_by_value(
        tf.round((values - val_min) * (127 / (val_max - val_min))), 0, 127)


def get_reshape_factor(knn_points):
    """Obtains the reshape factor from number of points to include in each
    localised group.

    Args:
        knn_points (int): number of points to include in each localised group.
            Must be a power of 2, from 4 to 512.

    Returns:
        int: the computed reshape factor.
    """
    if knn_points in [4, 8]:
        reshape_factor = 2
    elif knn_points in [16, 32]:
        reshape_factor = 4
    elif knn_points in [64, 128]:
        reshape_factor = 8
    elif knn_points in [256, 512]:
        reshape_factor = 16
    else:
        raise ValueError("'knn_points' argument must be a power of 2, from 4 " \
                         f"to 512 but received {knn_points}.")
    return reshape_factor


def pointnet_preproc(points, label, selected_points, knn_points):
    """ PointNet++ preprocessing function for sampling and grouping points.

    The preprocessing function is called for every point cloud to evaluate as
    it creates the input vector to our model from a point cloud.
    It has been empirically proven by the authors of the point cloud deep
    learning paper [PointNet (Qi et al., 2017)](https://arxiv.org/abs/1612.00593)
    that sampling a few points from the point cloud was enough to provide an
    input vector from which meaningful features could be extracted.

    In the point sampling and grouping operation:
    - a fixed number of points are selected from input points which defines the
      centroids of local regions.
    - the grouping process consists of identifying the set of points nearest to
      a centroid, and expressing their position relative to that centroid.

    Args:
        points (tf.Tensor): array representing a normalized point cloud.
        label (tf.Tensor): the labels.
        selected_points (int): number of points to process per sample.
        knn_points (int): number of neighboring points for the centroids to
            include in each localised group.

    Returns:
        tf.Tensor, tf.Tensor: the points preprocessed and labels.
    """

    input_points = points.shape[1]
    num_features = points.shape[-1]

    # The data loaded from file is normalized to the range [-1, 1]
    # However, for Akida, we need the inputs to be in the uint8 range (0, 255)
    # Also, the preprocessing function performs a substraction between point
    # coordinates, so some values are guaranteed to be negative at that point
    # in the pipeline.
    points = shift_to_7bit(points, -1.0, 1.0)

    # Random point selector
    # Simplification: selects same indices within a batch, and doesn't do
    # farthest point sampling
    sampler = tf.random.shuffle(tf.range(input_points),
                                seed=40)[:selected_points]
    inputs = tf.gather(points, sampler, axis=1, batch_dims=0)

    # Expects input of batch_size x sample_points x features
    # (with x, y, z, as the first 3 features)
    # batch_size x 1 x XYZ x sample_points
    temp1 = tf.expand_dims(tf.transpose(inputs[:, :, :3], perm=(0, 2, 1)), 1)
    # batch_size x selected_point x XYZ x 1
    temp2 = tf.expand_dims(inputs[:, :, :3], -1)
    points_dists = tf.reduce_sum(tf.square(temp1 - temp2), axis=2)
    # Note, LOWEST k, so top_k on negative values
    dists, points_idx = tf.math.top_k(-points_dists, k=knn_points, sorted=True)
    dists = -dists
    # dists and points_idx are batch_size x sample_points x knn
    # points_idx is size batch x selected_points x knn
    points_idx = tf.reshape(points_idx, [-1, selected_points * knn_points])
    # resized to batch x (selected_points * knn) x 1
    # Use that to gather from batch x all_points x features
    outpoints = tf.gather(inputs, points_idx, axis=1, batch_dims=1)
    outpoints = tf.reshape(outpoints,
                           [-1, selected_points, knn_points, num_features])
    # B x selected_points x 1 x 3
    points_center = tf.expand_dims(inputs[:, :selected_points, :3], axis=2)

    # Keep coordinates of original points
    subtraction_helper_l1 = tf.zeros_like(points_center)
    subtraction_helper_l2_to_end = tf.repeat(points_center,
                                             knn_points - 1,
                                             axis=2)
    subtraction_helper = tf.concat(
        [subtraction_helper_l1, subtraction_helper_l2_to_end], 2)

    if num_features > 3:
        sh = tf.shape(outpoints)
        subtraction_padding = tf.zeros((sh[0], sh[1], 1, sh[3] - 3))
        subtraction_helper = tf.concat(
            [subtraction_helper, subtraction_padding], -1)

    outpoints = (outpoints - subtraction_helper) / 127.0

    reshape_factor = get_reshape_factor(knn_points)

    # put the dimension to be folded last
    x = tf.transpose(outpoints, perm=[0, 3, 1, 2])
    # fold
    x = tf.reshape(x, [
        -1, num_features, selected_points * reshape_factor,
        int(knn_points / reshape_factor)
    ])
    # put dimensions back to original order
    points = tf.transpose(x, perm=[0, 2, 3, 1])
    return points, label



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Processing tools for YOLO data handling.
"""

import os

import xml.etree.ElementTree as et
import numpy as np
import tensorflow as tf


class BoundingBox:
    """ Utility class to represent a bounding box.

    The box is defined by its top left corner (x1, y1), bottom right corner
    (x2, y2), label, score and classes.
    """

    def __init__(self, x1, y1, x2, y2, score=-1, classes=None):
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        self.label = -1
        self.score = score
        self.classes = classes

    def __repr__(self):
        return "<BoundingBox({}, {}, {}, {}, {}, {}, {})>\n".format(
            self.x1, self.x2, self.y1, self.y2, self.get_label(),
            self.get_score(), self.classes)

    def get_label(self):
        """ Returns the label for this bounding box.

        Returns:
            Index of the label as an integer.
        """
        if self.label == -1:
            self.label = np.argmax(self.classes)

        return self.label

    def get_score(self):
        """ Returns the score for this bounding box.

        Returns:
            Confidence as a float.
        """
        if self.score == -1:
            self.score = self.classes[self.get_label()]
        return self.score

    def iou(self, other):
        """ Computes intersection over union ratio between this bounding box and
        another one.

        Args:
            other (BoundingBox): the other bounding box for IOU computation

        Returns:
            IOU value as a float
        """

        def _interval_overlap(interval_1, interval_2):
            x1, x2 = interval_1
            x3, x4 = interval_2

            if x3 < x1:
                if x4 < x1:
                    return 0
                return min(x2, x4) - x1
            if x2 < x3:
                return 0
            return min(x2, x4) - x3

        intersect_w = _interval_overlap([self.x1, self.x2],
                                        [other.x1, other.x2])
        intersect_h = _interval_overlap([self.y1, self.y2],
                                        [other.y1, other.y2])

        intersect = intersect_w * intersect_h

        w1, h1 = self.x2 - self.x1, self.y2 - self.y1
        w2, h2 = other.x2 - other.x1, other.y2 - other.y1

        union = w1 * h1 + w2 * h2 - intersect

        return float(intersect) / union


def load_image(image_path):
    """ Loads an image from a path.

    Args:
        image_path (string): full path of the image to load

    Returns:
        a Tensorflow image Tensor
    """
    raw_image = tf.io.read_file(image_path)
    return tf.image.decode_jpeg(raw_image, channels=3)


def preprocess_image(image_buffer, output_size):
    """ Preprocess an image for YOLO inference.

    Args:
        image_buffer (tf.Tensor): image to preprocess
        output_size (tuple): shape of the image after preprocessing

    Returns:
        A resized and normalized image as a Numpy array.
    """
    # Resize
    width = tf.constant(output_size[0])
    height = tf.constant(output_size[1])
    image = tf.compat.v1.image.resize(image_buffer, [height, width],
                                      method=tf.image.ResizeMethod.BILINEAR,
                                      align_corners=False)
    return image.numpy()


def decode_output(output,
                  anchors,
                  nb_classes,
                  obj_threshold=0.5,
                  nms_threshold=0.5):
    """ Decodes a YOLO model output.

    Args:
        output (tf.Tensor): model output to decode
        anchors (list): list of anchors boxes
        nb_classes (int): number of classes
        obj_threshold (float): confidence threshold for a box
        nms_threshold (float): non-maximal supression threshold

    Returns:
        List of `BoundingBox` objects
    """

    def _sigmoid(x):
        return 1. / (1. + np.exp(-x))

    def _softmax(x, axis=-1, t=-100.):
        x = x - np.max(x)

        if np.min(x) < t:
            x = x / np.min(x) * t

        e_x = np.exp(x)

        return e_x / e_x.sum(axis, keepdims=True)

    grid_h, grid_w, nb_box = output.shape[:3]

    boxes = []

    # decode the output by the network
    output[..., 4] = _sigmoid(output[..., 4])
    output[...,
           5:] = output[..., 4][..., np.newaxis] * _softmax(output[..., 5:])
    output[..., 5:] *= output[..., 5:] > obj_threshold

    for row in range(grid_h):
        for col in range(grid_w):
            for b in range(nb_box):
                # from 5th element onwards are class classes
                classes = output[row, col, b, 5:]

                if np.sum(classes) > 0:
                    # first 4 elements are x, y, w, and h
                    x, y, w, h = output[row, col, b, :4]

                    x = (col + _sigmoid(x)
                        ) / grid_w  # center position, unit: image width
                    y = (row + _sigmoid(y)
                        ) / grid_h  # center position, unit: image height
                    w = anchors[b][0] * np.exp(w) / grid_w  # unit: image width
                    h = anchors[b][1] * np.exp(h) / grid_h  # unit: image height

                    confidence = output[row, col, b, 4]

                    x1 = max(x - w / 2, 0)
                    y1 = max(y - h / 2, 0)
                    x2 = min(x + w / 2, grid_w)
                    y2 = min(y + h / 2, grid_h)

                    box = BoundingBox(x1, y1, x2, y2, confidence, classes)

                    boxes.append(box)

    # suppress non-maximal boxes
    for c in range(nb_classes):
        sorted_indices = list(
            reversed(np.argsort([box.classes[c] for box in boxes])))
        for ind, index_i in enumerate(sorted_indices):
            if boxes[index_i].classes[c] == 0:
                continue

            for j in range(ind + 1, len(sorted_indices)):
                index_j = sorted_indices[j]

                # filter out redundant boxes (same class and overlapping too
                # much)
                if (boxes[index_i].iou(boxes[index_j]) >= nms_threshold) and (
                        c == boxes[index_i].get_label()) and (
                            c == boxes[index_j].get_label()):
                    boxes[index_j].score = 0

    # remove the boxes which are less likely than a obj_threshold
    boxes = [box for box in boxes if box.get_score() > obj_threshold]

    return boxes


def parse_voc_annotations(gt_folder, image_folder, file_path, labels):
    """ Loads PASCAL-VOC data.

    Data is loaded using the groundtruth informations and stored in a
    dictionary.

    Args:
        gt_folder (str): path to the folder containing ground truth files
        image_folder (str): path to the folder containing the images
        file_path (str): file containing the list of files to parse
        labels (list): list of labels of interest
    Returns:
        dict: a dictionnary containing all data present in the ground truth file

    """

    def _build_list(file_path):
        """
        Builds a list of file names from a text file path.

        Args:
            file_path (string): path of the file containing the names
        Returns:
            list: list of file names
        """
        file_list = []
        with open(file_path, "r", encoding='UTF-8') as f:
            for line in f:
                file_list.append(line.rstrip())
        return file_list

    file_list = _build_list(file_path)

    all_data = []

    for f in file_list:
        tree = et.parse(os.path.join(gt_folder, f) + ".xml")

        data = {"boxes": []}

        for elem in tree.iter():
            if 'filename' in elem.tag:
                data['image_path'] = os.path.join(image_folder, elem.text)
            if 'size' in elem.tag:
                for attr in list(elem):
                    if 'width' in attr.tag:
                        width = int(attr.text)
                    if 'height' in attr.tag:
                        height = int(attr.text)
                    if 'depth' in attr.tag:
                        depth = int(attr.text)
                data['image_shape'] = (width, height, depth)

            if 'object' in elem.tag:
                box = {}

                for attr in list(elem):
                    if 'difficult' in attr.tag:
                        if int(attr.text) != 0:
                            box.clear()
                            break

                    if 'name' in attr.tag:
                        box['label'] = attr.text

                        if box['label'] not in labels:
                            box.clear()
                            break

                    if 'bndbox' in attr.tag:
                        for dim in list(attr):
                            if 'xmin' in dim.tag:
                                box['x1'] = int(round(float(dim.text)))
                            if 'ymin' in dim.tag:
                                box['y1'] = int(round(float(dim.text)))
                            if 'xmax' in dim.tag:
                                box['x2'] = int(round(float(dim.text)))
                            if 'ymax' in dim.tag:
                                box['y2'] = int(round(float(dim.text)))

                if len(box) == 5:
                    data["boxes"].append(box)

        if len(data["boxes"]) != 0:
            all_data.append(data)

    return all_data


def parse_widerface_annotations(gt_file, image_folder):
    """ Loads WiderFace data.

    Data is loaded using the groundtruth informations and stored in a
    dictionary.

    Args:
        gt_file (str): path to the ground truth file
        image_folder (str): path to the directory containing the images
    Returns:
        dict: a dictionnary containing all data present in the ground truth file

    """

    def _is_valid_box(w_box, h_box, w_img, h_img):
        box_area = w_box * h_box
        img_area = w_img * h_img
        return box_area >= img_area / 60.0

    all_data = []

    with open(gt_file, "r", encoding='UTF-8') as f:
        for line in f:
            file_path = line.rstrip()
            full_file_path = os.path.join(image_folder, str(file_path))

            if not os.path.isfile(full_file_path):
                raise RuntimeError(f"Not a file path : {full_file_path}")

            image = load_image(full_file_path)
            shape = (image.shape[1], image.shape[0], image.shape[2])

            data = {
                "image_path": full_file_path,
                "image_shape": shape,
                "boxes": []
            }

            nb_boxes = int(f.readline())
            if nb_boxes == 0:
                f.readline()

            for _ in range(nb_boxes):
                annotation = f.readline()
                values = annotation.split(' ')

                x1 = int(values[0])
                y1 = int(values[1])
                w = int(values[2])
                h = int(values[3])

                if _is_valid_box(w, h, shape[0], shape[1]):
                    box = {
                        "x1": x1,
                        "y1": y1,
                        "x2": x1 + w,
                        "y2": y1 + h,
                        "label": "face"
                    }
                    data["boxes"].append(box)

            if len(data["boxes"]) != 0:
                all_data.append(data)

        return all_data



# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Preprocessing tools for ModelNet40 data handling.
"""

import os
import glob
import numpy as np
import tensorflow as tf
import trimesh

from keras.utils.data_utils import get_file

from .pointnet_utils import pointnet_preproc


def parse_dataset(data_dir, num_points=1024):
    """ Parse the dataset.

    Args:
        data_dir (str): directory that contains the dataset.
        num_points (int, optional): number of points  with which mesh is sample.
            Defaults to 1024.

    Returns:
        tuple: tuple of train and test points with the corresponding labels.
    """
    print('Preparing Dataset. May take a few minutes')
    print('This is only done once for a given requested number of points')
    train_points = []
    train_labels = []
    test_points = []
    test_labels = []
    class_map = {}
    folders = glob.glob(os.path.join(data_dir, "*"))
    folders.sort()

    i = -1
    for folder in folders:
        if not folder.endswith('README.txt'):
            i += 1
            print("processing class: {}".format(os.path.basename(folder)))
            # store folder name with ID so we can retrieve later
            class_map[i] = os.path.basename(os.path.normpath(folder))
            # gather all files
            train_files = glob.glob(os.path.join(folder, "train/*"))
            test_files = glob.glob(os.path.join(folder, "test/*"))

            # Raw files have extremely varying amplitude ranges
            # Here, we normalise all to the range [-1, 1]
            for f in train_files:
                temp_points = np.array(trimesh.load(f).sample(num_points))
                temp_points /= np.amax(np.abs(temp_points))
                train_points.append(temp_points)
                train_labels.append(i)

            for f in test_files:
                temp_points = np.array(trimesh.load(f).sample(num_points))
                temp_points /= np.amax(np.abs(temp_points))
                test_points.append(temp_points)
                test_labels.append(i)

    return (
        np.array(train_points),
        np.array(test_points),
        np.array(train_labels),
        np.array(test_labels),
        class_map,
    )


def data_transformation(points, label):
    """ Data transformation function.

    Args:
        points (tf.Tensor): the points to which jitter is added.
        label (tf.Tensor): the labels.

    Returns:
        tf.Tensor, tf.Tensor: the points and labels for data transformation.
    """
    # jitter points
    points += tf.random.uniform(points.shape, -0.005, 0.005, tf.float64)
    # shuffle points
    points = tf.random.shuffle(points)
    return points, label


def get_modelnet_from_file(num_points, filename="ModelNet40.zip"):
    """ Load the ModelNet data from file.

    First parse through the ModelNet data folders. Each mesh is loaded and
    sampled into a point cloud before being added to a standard python list and
    converted to a `numpy` array. We also store the current enumerate index
    value as the object label and use a dictionary to recall this later.

    Args:
        num_points (int): number of points with which mesh is sample.
        filename (str): the dataset file to load if the npz file was not
            generated yet. Defaults to "ModelNet40.zip".

    Returns:
        np.array, np.array, np.array, np.array, dict: train set, train labels,
            test set, test labels as numpy arrays and dict containing class
            folder name.
    """

    base = os.path.basename(filename)
    short_filename = os.path.splitext(base)[0]

    datafile = short_filename + "_" + str(num_points) + "pts.npz"
    if os.path.exists(datafile):
        loadin = np.load(datafile, allow_pickle=True)
        train_points = loadin['train_points']
        train_labels = loadin['train_labels']
        test_points = loadin['test_points']
        test_labels = loadin['test_labels']
        class_map = loadin['class_map'].item()
    else:
        data_dir = filename

        # Join the test directory if needed
        is_exist = os.path.exists(data_dir)
        if not is_exist:
            data_dir = os.path.join("modelnet40", filename)
            is_exist = os.path.exists(data_dir)

        # Download dataset if not yet on local
        if not is_exist:
            original_url = "http://modelnet.cs.princeton.edu/" + filename
            ## Load dataset
            # For ModelNet40, 2 Gb to download, and 12K files to unzip,
            # so can take some time
            print("Checking for downloaded, unzipped data (one-time task)...")
            data_dir = get_file(
                filename,
                original_url,
                extract=True,
            )
            data_dir = os.path.join(os.path.dirname(data_dir), short_filename)

        # Then read in the data and store a fixed number of points per sample
        train_points, test_points, train_labels, test_labels, class_map = parse_dataset(
            data_dir, num_points)
        np.savez(datafile,
                 train_points=train_points,
                 train_labels=train_labels,
                 test_points=test_points,
                 test_labels=test_labels,
                 class_map=class_map)
    return train_points, train_labels, test_points, test_labels, class_map


def get_modelnet(train_points,
                 train_labels,
                 test_points,
                 test_labels,
                 batch_size,
                 selected_points=128,
                 knn_points=64):
    """ Obtains the ModelNet dataset.

    Args:
        train_points (numpy.array): train set.
        train_labels (numpy.array): train labels.
        test_points (numpy.array): test set.
        test_labels (numpy.array): test labels.
        batch_size (int): size of the batch.
        selected_points (int): num points to process per sample. Default is 512.
        knn_points (int): number of points to include in each localised group.
            Must be a power of 2, and ideally an integer square (so 64, or 16
            for a deliberately small network, or 256 for large).

    Returns:
        tf.data.Dataset, tf.data.Dataset: train and test point with data
            augmentation.
    """

    # Our data can now be read into a `tf.data.Dataset()` object. We set the
    # shuffle buffer size to the entire size of the dataset as prior to this the
    # data is ordered by class. Data augmentation is important when working with
    # point cloud data. We create an augmentation function to jitter and shuffle
    # the train dataset.

    train_dataset = tf.data.Dataset.from_tensor_slices(
        (train_points, train_labels))
    test_dataset = tf.data.Dataset.from_tensor_slices(
        (test_points, test_labels))

    if len(train_points) > 0:
        train_dataset = train_dataset.shuffle(
            len(train_points)).map(data_transformation).batch(batch_size)
        train_dataset = \
            train_dataset.map(lambda points,
                              label: pointnet_preproc(points, label,
                                                   selected_points, knn_points),
                              num_parallel_calls=tf.data.experimental.AUTOTUNE)

    test_dataset = test_dataset.batch(batch_size)
    test_dataset = \
        test_dataset.map(lambda points,
                         label: pointnet_preproc(points, label, selected_points,
                                              knn_points),
                         num_parallel_calls=tf.data.experimental.AUTOTUNE)

    return train_dataset, test_dataset



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Preprocessing tools for face recognition: allows to detect faces and their
associated landmarks and align the faces before processing.

Based on:
    - MTCNN https://arxiv.org/ftp/arxiv/papers/1604/1604.02878.pdf
        title: Joint face detection and alignment using multitask cascaded
            convolutional networks.
        author: Zhang, K., Zhang, Z., Li, Z., and Qiao, Y. (2016)
    - matlab cp2tform python implementation to mimic SphereFace original code:
        https://github.com/clcarwin/sphereface_pytorch/blob/master/matlab_cp2tform.py
"""

import cv2
import numpy as np

from mtcnn import MTCNN
from numpy.linalg import inv, norm, lstsq


def _find_non_reflective_similarity(uv, xy):
    """
    Find Non-reflective Similarity Transform Matrix 'trans':
            u = uv[:, 0]
            v = uv[:, 1]
            x = xy[:, 0]
            y = xy[:, 1]
            [x, y, 1] = [u, v, 1] * trans

    Args:
        uv (np.array): source points each row is a pair of coordinates (x, y)
        xy (np.array): each row is a pair of inverse-transformed

    Returns:
        np.array: transform matrix from uv to xy
    """

    M = xy.shape[0]
    x = xy[:, 0].reshape((-1, 1))  # use reshape to keep a column vector
    y = xy[:, 1].reshape((-1, 1))

    tmp1 = np.hstack((x, y, np.ones((M, 1)), np.zeros((M, 1))))
    tmp2 = np.hstack((y, -x, np.zeros((M, 1)), np.ones((M, 1))))
    X = np.vstack((tmp1, tmp2))

    u = uv[:, 0].reshape((-1, 1))  # use reshape to keep a column vector
    v = uv[:, 1].reshape((-1, 1))
    U = np.vstack((u, v))

    # We know that X * r = U
    r, _, _, _ = lstsq(X, U)
    r = np.squeeze(r)

    sc = r[0]
    ss = r[1]
    tx = r[2]
    ty = r[3]

    Tinv = np.array([[sc, -ss, 0], [ss, sc, 0], [tx, ty, 1]])

    T = inv(Tinv)
    T[:, 2] = np.array([0, 0, 1])

    return T


def _get_similarity_transform(src_pts, dst_pts):
    """
    Find Reflective Similarity Transform Matrix which can be directly used by
    cv2.warpAffine().

    Args:
        src_pts (np.array): source points
        dst_pts (np.array): destination points

    Returns:
        np.array: transform matrix from src_pts to dst_pts
    """

    def tformfwd(trans, uv):
        """ Apply affine transform 'trans' to uv
        """
        uv = np.hstack((uv, np.ones((uv.shape[0], 1))))
        xy = np.dot(uv, trans)
        return xy[:, 0:-1]

    # Solve for trans1
    trans1 = _find_non_reflective_similarity(src_pts, dst_pts)

    # Solve for trans2

    # manually reflect the xy data across the Y-axis
    xyR = dst_pts
    xyR[:, 0] = -1 * xyR[:, 0]

    trans2r = _find_non_reflective_similarity(src_pts, xyR)

    # manually reflect the tform to undo the reflection done on xyR
    TreflectY = np.array([[-1, 0, 0], [0, 1, 0], [0, 0, 1]])

    trans2 = np.dot(trans2r, TreflectY)

    # Figure out if trans1 or trans2 is better
    xy1 = tformfwd(trans1, src_pts)
    norm1 = norm(xy1 - dst_pts)

    xy2 = tformfwd(trans2, src_pts)
    norm2 = norm(xy2 - dst_pts)

    if norm1 <= norm2:
        cv2_trans = trans1[:, 0:2].T
    else:
        cv2_trans = trans2[:, 0:2].T

    return cv2_trans


def preprocess_face(image_path, dst_shape=(96, 112)):
    """ Preprocess an image for face recognition by detecting the face and face
    landmarks in it before aligning the face and croping to the given size.

    Args:
        image_path (str): file path to the image

    Returns:
        np.array: the preprocessed image
    """

    # Define preprocessing constant
    ref_points = np.array([(30.2946, 51.6963), (65.5318, 51.5014),
                           (48.0252, 71.7366), (33.5493, 92.3655),
                           (62.7299, 92.2041)]).astype(np.float32)

    # Read image file
    base_image = cv2.imread(image_path)

    # Build MTCNN detector and detect faces
    detector = MTCNN()
    landmarks = detector.detect_faces(base_image)

    # Return None if no face detected
    if len(landmarks) == 0:
        return None

    # Only consider first face keypoints
    detected_points = np.array(list(landmarks[0]["keypoints"].values())).astype(
        np.float32)

    # Map keypoints to reference and apply transformation
    transform = _get_similarity_transform(detected_points, ref_points)
    return cv2.warpAffine(base_image, transform, dst_shape)
